"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d.core import TextureAttrib, DepthWriteAttrib, TransparencyAttrib
from panda3d.core import ColorAttrib, ModelNode, NodePath, Texture, TextureStage

from direct.actor.Actor import Actor

from panda3d_gemstone.framework.resource import Resource, ExportProperty, ResourceManager
from panda3d_gemstone.io.file_system import path_exists
from panda3d_gemstone.framework.internal_object import InternalObject

from panda3d_gemstone.engine import runtime
from panda3d_gemstone.engine.model_utilities import get_all_geom_nodes, get_all_geom_states
from panda3d_gemstone.engine.loader import attempt_load_model, attempt_unload_model

class Model(Resource, InternalObject):
    """
    Represents a basic model in the Gemstone framework for game and editor use
    """

    def __init__(self, config_path: str, force_export: bool = False, load_immediate: bool = True):
        self._disable_setting_render_state = False
        InternalObject.__init__(self)
        Resource.__init__(self, config_path, force_export, load_immediate)

    @property
    def disable_setting_render_state(self) -> bool:
        """
        """

        return self.get_disable_setting_render()

    @disable_setting_render_state.setter
    def disable_setting_render_state(self, state) -> None:
        """
        """

        self.set_disable_setting_render_state(state)

    def set_disable_setting_render_state(self, state: bool) -> None:
        """
        """

        self._disable_setting_render_state = state

    def get_disable_setting_render(self):
        """
        """

        return self._disable_setting_render_state

    def prepare_as_rp_scene(self) -> None:
        """
        Prepares the model as a render pipeline scene
        """

        if not runtime.has_render_pipeline():
            self.notify.warning('Attempted to prepare as rp scene. RenderPipeline not found.')
            return

        runtime.render_pipeline.prepare_scene(self)

    def fixup_render_state(self) -> None:
        """
        """

        if self._disable_setting_render_state:
            return

        for geom_node in get_all_geom_nodes(self):
            for i in range(geom_node.get_num_geoms()):
                state = geom_node.get_geom_state(i)
                texture_attrib = state.get_attrib(TextureAttrib.get_class_type())

                if not texture_attrib:
                    continue

                texture = texture_attrib.get_texture()
                if not texture:
                    continue

                filename = texture.get_filename().get_basename_wo_extension()
                transparency_attrib = None
                color_attrib = None
                depth_write_attr = None

                if filename.find('_plant') >= 0:
                    transparency_attrib = TransparencyAttrib.make(TransparencyAttrib.MDual)
                
                if filename.find('_mask') >= 0:
                    transparency_attrib = TransparencyAttrib.make(TransparencyAttrib.MBinary)

                if filename.find('_shadow') >= 0:
                    depth_write_attr = DepthWriteAttrib.make(DepthWriteAttrib.MOff)
                    transparency_attrib = TransparencyAttrib.make(TransparencyAttrib.MAlpha)

                if filename.find('_bright') >= 0:
                    color_attrib = ColorAttrib.make_off()

                for attrib in [transparency_attrib, color_attrib, depth_write_attr]:
                    if attrib:
                        geom_node.set_geom_state(i, geom_node.get_geom_state(i).add_attrib(attrib))
 
class StaticModel(Model, NodePath):
    """
    Represents a static model in the gemstone framework
    """

    def __init__(self, config_path: str, force_export: bool = False, load_immediate: bool = True):
        self._optimize_geometries = True
        self._generate_txo = False
        self._generate_mipmaps = False
        self._override_texture = None
        self._override_texture_name = None
        self._models = []
        self._model_data = {}
        self._pending_loads = 0

        NodePath.__init__(self, self.__class__.__name__)
        Model.__init__(self, config_path, force_export, load_immediate)

    @property
    def optimized_geometries(self) -> bool:
        """
        Returns the optimized geometries flag as a property
        """

        return self.get_optimize_geometries()

    @optimized_geometries.setter
    def optimized_geometries(self, value) -> None:
        """
        Sets the optimized geometries flag as a setter
        """

        self.set_optimize_geometries(value)

    def destroy(self) -> None:
        """
        Destroys the static model instance
        """
        
        Model.destroy(self)
        self.remove_node()

    def load_model_data(self, data: dict) -> None:
        """
        Loads a data dictionary into the static model
        """

        self._model_data = data

    def set_optimize_geometries(self, optimized_geometries: bool) -> None:
        """
        Sets the optimized geometries flag
        """

        self._optimize_geometries = optimized_geometries        

    def get_optimize_geometries(self) -> bool:
        """
        Returns the optimized geometries flag
        """

        return self._optimize_geometries

    def set_generate_txo(self, generate_txo: bool) -> None:
        """
        """

        self._generate_txo = generate_txo

    def get_generate_txo(self) -> bool:
        """
        """

        return self._generate_txo

    def set_generate_mipmaps(self, generate_mipmaps: bool) -> None:
        """
        """

        self._generate_mipmaps = generate_mipmaps

    def get_generate_mipmaps(self) -> None:
        """
        """

        return self._generate_mipmaps

    def set_override_texture(self, override_texture: object) -> None:
        """
        """

        self._override_texture = override_texture

    def get_override_texture(self) -> object:
        """
        """

        return self._override_texture

    def _generate_texture_options(self) -> dict:
        """
        Generates the texture export options
        """

        egg2bam_options = {}
        if self._generate_txo:
            egg2bam_options['-txo'] = ''
        if self._generate_mipmaps:
            egg2bam_options['-mipmaps'] = ''

        return egg2bam_options

    def export(self, force_export: bool) -> None:
        """
        Exports the static model to disk
        """

        self._models = []
        texture_options = self._generate_texture_options()
        for part, value in list(self._model_data.items()):
            export = ExportProperty(part, value, 'Export')
            export.options['-keep-uvs'] = ''
            self.add_export_step(export)
            #self.extend_export_steps(export.generate_post_egg2bampz(texture_options))
            self._models.append(self.get_export_step(-1).output_filename)

        Model.export(self, force_export)

    def _load(self, force_export: bool) -> bool:
        """
        """

        # Verify we are not already loaded
        if self.loaded:
            return False

        success = Model._load(self, force_export)
        if not success:
            return False

        if len(self._models) == 0:
            self.notify.warning('Failed to load %s. No models to load' % self.__class__.__name__)
            return False

        loader_options = ResourceManager.get_singleton().get_loader_options()
        for model_name in self._models:
            model = self.__load_model(model_name, loaderOptions=loader_options)

        return True

    def do_override_texture(self, texture_name: str) -> None:
        """
        """

        try:
            texture = runtime.loader.load_texture(texture_name)
            if texture:
                texture.set_magfilter(Texture.FTLinear)
                texture.set_minfilter(Texture.FTLinearMipmapLinear)

                for geom_node in get_all_geom_nodes(self):
                    for i in range(geom_node.get_num_geoms):
                        state = geom_node.get_geom_state(i)
                        geom_node.set_geom_state(i, state.add_attrib(TextureAttrib.make(texture)))
        except IOError:
            self.notify.warning('%s.do_override_texture: Failed to load texture "%s"' % (
                self.__class__.__name__, texture_name))

    def __process_model(self, model) -> None:
        """
        Processes a newly loaded model in async via a callback
        """

        if not model:
            self.notify.error('Failed to load model for object %s.' % self.__class__.__name__)
            return

        if self.is_empty():
            self.notify.warning('Failed to load model for object %s. StaticModel is empty' % (
                self.__class__.__name__))

            return

        if not self._optimize_geometries:
            model.reparent_to(self)
        else:
            model.get_children().reparent_to(self)
        self._pending_loads -= 1

        # Check if we are done loading
        if self._pending_loads <= 0:
            self.__finish_load()

    def __finish_load(self) -> None:
        """
        Called on model loading complete
        """

        if self._optimize_geometries:
            self.flatten_strong()

        if self._override_texture:
            self.do_override_texture(self._override_texture)

        self.fixup_render_state()

    def __load_model(self, model_name: str, **args) -> None:
        """
        Loads the requested model from the VFS
        """

        try:
            success = attempt_load_model(model_name, callback=self.__process_model, **args)
            if not success:
                self.notify.warning('Failed to load model (%s) for %s.' % (
                    model_name, self.__class__.__name__))
            else:
                self._pending_loads += 1
        except IOError:
            pass

    def _unload(self) -> bool:
        """
        Unloads the StaticModel object
        """

        if not self.loaded:
            return False
        
        for i in range(self.get_num_children() - 1, -1, -1):
            self.get_child(i).remove_node()

        for model_name in self._models:
            success = attempt_unload_model(model_name)
            if not success:
                self.notify.warning('%s.__unload: Failed to unload model (%s).' % (
                    self.__class__.__name__, model_name))

class TextureModel(StaticModel):
    """
    Represents a texture model in the Gemstone framework
    """

    def __init__(self, config_path: str, force_export: bool = False):
        self.__generate_ico = False
        self.__width = 0
        self.__height = 0
        self.__depth = 0
        super().__init__(config_path, force_export)
        self.__cache_width_and_height()

    @property
    def generate_ico(self) -> bool:
        """
        Returns the model's generate ico attribute
        as a property
        """

        return self.get_generate_ico()

    @generate_ico.setter
    def generate_ico(self, generate_ico: bool) -> None:
        """
        Sets the model's generate ico attribute
        as a setter
        """

        self.set_generate_ico(generate_ico)

    @property
    def width(self) -> int:
        """
        Returns this model's cached width
        as a property
        """

        return self.get_width()

    @property
    def height(self) -> int:
        """
        Returns this model's cached height
        as a property
        """

        return self.get_height()

    @property
    def depth(self) -> int:
        """
        Returns this model's cached depth
        as a property
        """

        return self.get_depth()

    def set_generate_ico(self, generate_ico: bool) -> None:
        """
        Sets the texture model's generate ico attribute
        used in exporting
        """

        self.__generate_ico = generate_ico

    def get_generate_ico(self) -> bool:
        """
        Returns the texture model's generate ico attribute
        """

        return self.__generate_ico

    def export(self, force_export: bool) -> None:
        """
        Exports the TextureModel to disk
        """

        texture_options = self._generate_texture_options()
        for part, value in list(self._model_data.items()):
            export = ExportProperty(part, value, 'TextureCards')
            self.add_export_step(export)
            self.extend_export_steps(export.generate_post_bam2pz(texture_options))
            
            if self.__generate_ico:
                export = ExportProperty(part, value, 'Png2Ico')
                self.add_export_step(export)

            self._models.append(self.get_export_step(-1).output_filename)

        Resource.export(self, force_export)

    def __cache_width_and_height(self) -> None:
        """
        Caches the TextureModels width and height
        of its geom textures
        """

        self.__width = 0
        self.__height = 0
        self.__depth = 0

        for geom_state in get_all_geom_states(self):
            texture_attrib = geom_state.get_attrib(TextureAttrib.get_class_type())

            if not texture_attrib:
                continue

            texture = texture_attrib.get_texture()
            if not texture:
                continue

            self.__width = max(self.__width, texture.get_x_size())
            self.__height = max(self.__height, texture.get_y_size())
            self.__depth = max(self.__depth, texture.get_z_size())

    def get_width(self) -> int:
        """
        Returns the TextureModel's cached width
        """

        return self.__width
    
    def get_height(self) -> int:
        """
        Returns the TextureModel's cached height
        """

        return self.__height

    def get_depth(self) -> int:
        """
        Returns the TextureModel's cached depth
        """

        return self.__depth
    
class AnimatedModel(Model, Actor):
    """
    Represents an animated model in the Gemstone framework
    """

    def __init__(self, config_path: str, force_export: bool = False, have_anim_in_node: bool = True):
        self.__have_anim_in_node = have_anim_in_node
        self.__optimize_geometries = True
        self.__optimize_animations = False
        self.__generate_txo = False
        self.__generate_mipmaps = False
        self.__include_animation = False
        self.__models = []
        self.__load_on_demand = False
        self.__animations = []
        self.__animation_blend = False
        self.__frame_blend = False

        Actor.__init__(self)
        Model.__init__(self, config_path, force_export)

        resource_mgr = ResourceManager.get_singleton()
        if resource_mgr and hasattr(self, 'model_loader_options'):
            self.model_loader_options.set_flags(self.model_loader_options.get_flags() | resource_manager.get_loader_options().get_flags())
        
    def destroy(self) -> None:
        """
        Destroys the animated model
        """

        self.cleanup()
        self.remove_node()

    def np_copy_to(self, destination: object, sort: int = 0) -> object:
        """
        Copys the nodepath to a new destination with an optional sort
        """

        return NodePath.copy_to(self, destination, sort)

    def make_copy_to(self, destination: object, callback: object = None, extra_args: list = [], task_chain_name: str = None) -> None:
        """
        """

        destination.copyActor(other=self, overrwrite=False, callback=callback, extraArgs=extra_args, taskChainName=task_chain_name)
        destination.__optimize_geometries = self.__optimize_geometries
        destination.__optimize_animations = self.__optimize_animations
        destination.__generate_txo = self.__generate_txo
        destination.__generate_mipmaps = self.__generate_mipmaps
        destination.__include_animation = self.__include_animation
        destination.__models = copy.copy(self.__models)
        destination.__model_data = copy.copy(self.__model_data)
        destination.__load_on_demand = self.__load_on_demand
        destination.__animations = copy.copy(self.__animations)
        destination.__actions = copy.copy(self.__actions)
        destination.__emotes = copy.copy(self.__emotes)
        destination.__animation_data =copy.copy(self.__animation_data)
        destination.path = self.path
        destination.sectoin = self.section

    def load_model_data(self, data: dict) -> None:
        """
        Loads the model data into memory
        """

        self.__model_data = data

    def load_animation_data(self, data: dict) -> None:
        """
        Loads the animation data into memory
        """

        self.__animation_data = data

    def load_action_data(self, data: dict) -> None:
        """
        Loads the action data into memory
        """

        self.__actions = data

    def load_emote_data(self, data: dict) -> None:
        """
        Loads the emote data into memory
        """

        self.__emotes = data

    def get_actions(self) -> list:
        """
        Returns the available actions
        """

        return list(self.__actions.keys())

    def get_emotes(self) -> list:
        """
        Returns the available emotes
        """

        return list(self.__emotes.keys())

    def set_animation_blend(self, state: bool) -> None:
        """
        Sets the animated model's animation blend value
        """

        self.__animation_blend = state

    def set_frame_blend(self, state: bool) -> None:
        """
        Sets the animated model's frame blend value
        """

        self.__frame_blend = state
 
    def set_load_on_demand(self, load_on_demand: bool) -> None:
        """
        Sets the animated model's load on demand value
        """

        self.__load_on_demand = load_on_demand

    def set_optimize_geometries(self, optimized_geometries: bool) -> None:
        """
        Sets the aniamted model's optimize geometry flag for auto flattening
        """

        self.__optimize_geometries = optimized_geometries

    def get_optimize_geometries(self) -> bool:
        """
        """

        return self.__optimize_geometries

    def set_optimize_animations(self, optimize_animations: bool) -> None:
        """
        """

        self.__optimize_animations = optimize_animations

    def get_optimize_animations(self) -> bool:
        """
        """

        return self.__optimize_animations

    def set_generate_tx_(self, generate_txo: bool) -> None:
        """
        """

        self.__generate_txo = generate_txo

    def get_generate_txo(self) -> bool:
        """
        """

        return self.__generate_txo

    def set_generate_mipmaps(self, generate_mipmaps: bool) -> None:
        """
        """

        self.__generate_mipmaps = generate_mipmaps

    def get_generate_mipmaps(self) -> bool:
        """
        """

        return self.__generate_mipmaps

    def set_include_animation(self, include_animation: bool) -> None:
        """
        """

        self.__include_animation = include_animation

    def get_include_animation(self) -> bool:
        """
        """

        return self.__include_animation

    def _generate_texture_options(self) -> dict:
        """
        Returns the model's texture export options
        """

        egg2bam_options = {}
        if self.__generate_txo:
            egg2bam_options['-txo'] = ''

        if self.__generate_mipmaps:
            egg2bam_pptions['-mipmap'] = ''
        
        return egg2bam_options

    def export(self, force_export: bool) -> None:
        """
        Performs export operations on the AnimatedModel instance
        """

        texture_options = self._generate_texture_options()
        for part, value in list(self.__model_data.items()):
            export = ExportProperty(part, value, 'Export')
            if self.__include_animation:
                export.options['-a'] = 'both'
                export.options['-cn'] = 'idle'
            else:
                export.options['-a'] = 'model'
            export.options['-keep-uvs'] = ''
            self.add_export_step(export)
            
            previous_step = export
            if not self.__optimize_geometries:
                optimize = ExportProperty()
                optimize.input_filename = previous_step.output_filename
                optimize.output_filename = optimize.input_filename
                optimize.options['-flag'] = '"*"'
                optimize.options['-keepall']  = ''
                optimize.converter = 'OptChar'

                self.add_export_step(optimize)
                previous_step = optimize

            if self.__optimize_animations:
                self.notify.warning('Optimizing animations on export is not currently supported')

            self.extend_export_steps(export.generate_post_bam2pz())
            self.__models.append(self.get_export_step(-1).output_filename)

        for part, value in list(self.__animation_data.items()):
            export = ExportProperty(part, value, 'Export')
            export.options['-a'] = 'chan'
            self.add_export_step(export)
            self.extend_export_steps(export.generate_post_bam2pz())
            self.__animations.append(self.get_export_step(-1).output_filename)

        Model.export(self, force_export)

    def _load_model(self, model_name: str) -> None:
        """
        Loads the AnimatedModel instance from disk
        """

        self.notify.debug('Loading %s "%s"...' % (self.__class__.__name__, model_name))
        try:
            if path_exists(model_name + '.bam.pz'):
                self.load_model(model_name + '.bam.pz', autoBindAnims=self.__have_anim_in_node)
            elif path_exists(model_name + '.bam'):
                self.load_model(model_name + '.bam', autoBindAnims=self.__have_anim_in_node)
            elif path_exists(model_name + '.egg'):
                self.load_model(model_name + '.egg', autoBindAnims=self.__have_anim_in_node)    
            else:
                self.notify.warning('Failed to load model. Missing model part file "%s"' % model_name)
        except Exception:
            self.notify.warning('Failed to load model. Missing model part file "%s"' % model_name)

    def __get_all_animation_items(self) -> list:
        """
        Returns all configured animations in the AnimatedModel instance
        """

        return list(self.__actions.items()) + list(self.__emotes.items())

    def _load(self, force_export: bool) -> bool:
        """
        Loads the animated model into memory
        """

        if not Model._load(self, force_export):
            self.notify.error('Failed to load %s. Base load failed' % (
                self.__class__.__name__))
            return False

        if not len(self.__models):
            self.notify.error('Failed to load %s. No models to load' % (
                self.__class__.__name__))
            
            return False

        for part in self.__models:
            self._load_model(part)

        animation_dict = {}
        animation_items = self.__get_all_animation_items()
        if not len(animation_items):
            self.notify.warning('No animation items loaded into AnimatedModel')

        for action_name, action_file in animation_items:
            found = False

            for animation_name in self.__animations:
                if action_file == animation_name:
                    suffix = None
                    if path_exists(action_file + '.bam.pz'):
                        suffix = '.bam.pz'
                    elif path_exists(action_file + '.bam'):
                        suffix = '.bam'
                    elif path_exists(action_file + '.egg'):
                        suffix = '.egg'

                    if suffix:
                        animation_dict[action_name] = action_file + suffix
                        found = True

            if not found:
                self.notify.error('Missing animation file "%s" for action "%s"' % (action_file, action_name))

        self.set_blend(
            animBlend=True, #self.__animation_blend, #TODO: fix order of operations for the config
            frameBlend=self.__frame_blend)

        self.load_anims(animation_dict)
        if not self.__load_on_demand:
            try:
                for animation in animation_dict:
                    self.bind_anim(animation) 
            except KeyError as e:
                self.notify.warning('Failed to bind animations, Required part missing: %s' % str(e))

        self.fixup_render_state()
        return True

    def _unload(self) -> bool:
        """
        Unloads the AnimatedModel from memory
        """

        if Model._unload(self):
            children = self.get_geom_node().get_children()
            children.detach()
            for part in self.__models:
                self.remove_part(part)
                runtime.loader.unload_model(part + '.bam.pz')

            self.unload_anims(self.__actions)
            self.unload_anims(self.__emotes)

            return True

        return False


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import sys
import copy
import configparser
import pickle

from panda3d_gemstone.logging.utilities import get_notify_category
from panda3d_gemstone.engine import runtime, prc

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.io.file_system import get_file_date
from panda3d_gemstone.framework.cast import cast
from panda3d_gemstone.framework.utilities import get_snake_case

from panda3d.core import ConfigVariableBool, VirtualFileSystem
from panda3d.core import Filename

_config_notify = get_notify_category('configurable')

class BaseConfigurableCache(InternalObject):
    """
    Base class for all configurable cache handlers
    """

class VFSConfigurableCache(BaseConfigurableCache):
    """
    Caching object for the configuration files
    loaded from the Panda3D virtual file
    system
    """

    vfs = VirtualFileSystem.get_global_ptr()

    def __init__(self):
        super().__init__()
        self.__files = {}

    def read(self, filepath: str, cls: object) -> object:
        """
        Reads the requested file from the cache
        """

        filepath_dat = OSConfigurableCache.get_cache_filename(filepath)
        if filepath not in self.__files:
            filepath_dat_name = Filename(filepath_dat)
            fh = self.vfs.get_file(filepath_dat_name)
            if fh:
                reader = pickle.loads(fh.read_file(True))
                self.__files[filepath] = reader
            fh = None

        if filepath not in self.__files:
            self.notify.warning('Failed to read file: %s. Not found' % filepath)
            return

        return self.__files[filepath]

    def get_notify_name(self) -> str:
        """
        Returns this object's custom notifier name
        """

        return 'vfs-configurable-cache'

class RawOSConfigurableCache(BaseConfigurableCache):
    """
    Caching object for configuration files loaded from
    the operation system's IO without using binary dat
    files
    """

    def __init__(self):
        super().__init__()
        self._files = {}

    def __read_file(self, filepath: str, clsName: str) -> configparser.ConfigParser:
        """
        """

        reader = configparser.ConfigParser()
        reader.optionxform = str
        self.notify.info('Reading config: %s' % filepath)

        try:
            if not reader.read(filepath):
                self.notify.warning('Failed to load ini file "%s". Path does not exist' % (filepath))
                return None
        except Exception as e:
            self.notify.error('Failed to read ini file "%s": %s' % (filepath, str(e)))
            return None

        return reader

    def read(self, filepath: str, cls) -> object:
        """
        Reads the requested file from the cache
        """

        # Add the file to the cache if not already added
        if filepath not in self._files:
            reader = self.__read_file(filepath, cls.__name__)
            self._files[filepath] = (reader, get_file_date(filepath))

        return self._files[filepath][0]

    def get_notify_name(self) -> str:
        """
        Returns this object's custom notifier name
        """

        return 'raw-os-configurable-cache'

class OSConfigurableCache(BaseConfigurableCache):
    """
    Caching object for configuration files loaded from the
    operation system's IO 
    """

    def __init__(self):
        super().__init__()
        self._files = {}
        self._use_dat = ConfigVariableBool('gs-use-dat', False).value

    @staticmethod
    def get_cache_filename(filepath: str) -> str:
        """
        Returns the cache filename for the requested file path
        """

        output, ext = os.path.splitext(filepath)
        output += '.dat'

        return output

    def __read_file(self, filepath: str, clsName: str) -> configparser.ConfigParser:
        """
        """

        reader = configparser.ConfigParser()
        reader.optionxform = str
        self.notify.info('Reading config: %s' % filepath)

        try:
            if not reader.read(filepath):
                self.notify.warning('Failed to load ini file "%s". Path does not exist' % (filepath))
                return None
        except Exception as e:
            self.notify.error('Failed to read ini file "%s": %s' % (filepath, str(e)))
            return None

        return reader

    def read(self, filepath: str, cls) -> object:
        """
        Reads the requested file from the cache
        """

        dat_file = OSConfigurableCache.get_cache_filename(filepath)
        if get_file_date(filepath) > get_file_date(dat_file):

            # Create a new dat file if it does not exist
            try:
                reader = self.__read_file(filepath, cls.__name__)
                if reader:
                    fh = open(dat_file, 'wb')
                    fh.write(pickle.dumps(
                        reader,
                        protocol=pickle.HIGHEST_PROTOCOL))
                    fh.close()
            except IOError:
                pass

        # Add the file to the cache if not already added
        if filepath not in self._files:
            if get_file_date(dat_file) and self._use_dat:
                fh = open(dat_file, 'rb')
                reader = pickle.load(fh.read())
                fh.close()

                self._files[filepath] = (reader, get_file_date(dat_file))
            else:
                reader = self.__read_file(filepath, cls.__name__)
                self._files[filepath] = (reader, get_file_date(dat_file))
        elif get_file_date(dat_file) > self._files[filepath][1]:
            reader = pickle.load(open(dat_file, 'rb'))
            self._files[filepath] = (reader, get_file_date(dat_file))

        if filepath not in self._files:
            self.notify.warning('Failed to read file: %s. Not found' % filepath)
            return

        return self._files[filepath][0]

    def get_notify_name(self) -> str:
        """
        Returns this object's custom notifier name
        """

        return 'os-configurable-cache'

class Configurable(object):
    """
    Represents a configurable object in the Gemstone framework
    """

    __include_str = '__include__'
    __cache = OSConfigurableCache()

    def __init__(self, path: str, section: str = 'Configuration', auto_configure: bool = False, fixup_include_path_func: object = None, override_cache: object = None):
        self.__config = None
        self._included_config = None
        self._fixup_include_path_func = fixup_include_path_func
        self.__want_config_warnings = prc.get_prc_bool('gs-config-warnings', False)
        
        if not hasattr(self, 'path'):
            self.path = path
            self.section = section

        self.auto_configure = auto_configure
        self.configuration = {}
        self._override_cache = override_cache

        if path:
            self.load(path, section)

    @property
    def config(self) -> object:
        """
        Returns this configurables working config value
        """

        self.__config

    def __prepare_data(self, data: dict) -> dict:
        """
        Prepares the incoming data
        by casting it to its respective type
        """

        new_data = {}
        for key, value in data:
            new_data[key] = cast(value)

        return new_data

    def __fixup_include_config_path(self, config_path: str) -> str:
        """
        Runs the fixup include path function on the config path 
        if provided on construction
        """

        if not self._fixup_include_path_func:
            return config_path
        
        return self._fixup_include_path_func(config_path)

    def load(self, opath = None, osection = None) -> None:
        """
        Loads the configuration from the applications
        file system into memory
        """

        # Load the appropriate cache for the configuration
        # load operation
        if self._override_cache is not None:
            cache = self._override_cache
        else:
            cache = Configurable.__cache

        # Read the data from the cache and begin processing
        self.__config = cache.read(opath or self.path, self.__class__)
        if not self.__config:
            return

        processed_section = []
        main_section = osection or self.section

        if self.__config.has_section(main_section):
            data = self.__config.items(main_section)

            # Process include path if present
            include_path = dict(data).get(Configurable.__include_str)
            if include_path:
                included_configurable = Configurable(
                    path=self.__fixup_include_config_path(include_path),
                    fixup_include_path_func=self._fixup_include_path_func)
                self._included_config = included_configurable.config
                self.configuration = copy.copy(included_configurable.configuration)
                processed_section.append(main_section)

            # Populate section data in configuration dict
            for key, value in data:
                if key != Configurable.__include_str:
                    self.configuration[key] = cast(value) 

        # Load all sections in our configuration object
        for section in self.__config.sections():
            if section != main_section:
                data = self.__config.items(section)
                if self._included_config and self._included_config.has_section(section):
                    included_data = self._included_config.items(section)
                    included_data.extend(data)
                    data = included_data

                    self.__config.remove_section(section)
                    self.__copy_section(section, data)
                    processed_section.append(section)
                
                self.__load_section(section, data)
        
        # Process any included configurations
        if self._included_config:
            for section in self._included_config.sections():
                if section not in processed_section:
                    data = self._included_config.items(section)
                    self.__load_section(section, data)
                    self.__copy_section(section, data)

        self._included_config = None
        self._fixup_include_path_func = None

    def __copy_section(self, section: str, data: object) -> None:
        """
        Copys a section of data to the current configuration
        object instance
        """

        temp = dict(data)
        self.__config.add_section(section)

        # Population section with data
        for key, value in temp.items():
            self.__config.set(section, key, value)

    def __get_loader_name(self, section: str) -> str:
        """
        Returns the loader name for the
        requested config section
        """
        
        snake_case = get_snake_case(section)
        return 'load_%s_data' % snake_case

    def __load_section(self, section: str, data: object) -> None:
        """
        Loads the requested section into the
        configurable object
        """

        # Retrieve the loader function if present
        loader_name = self.__get_loader_name(section)

        # Process the section's data
        if hasattr(self, loader_name):
            loader = getattr(self, loader_name)
            loader(self.__prepare_data(data))
        else:
            if self.__want_config_warnings:
                _config_notify.warning('Failed to load section for %s. Section loader "%s" does not exist' % (
                    self.__class__.__name__, section))
            self.load_data(section, self.__prepare_data)

    def __get_snake_case(self, label: str) -> str:
        """
        Returns the label as a snake case variant
        """

        return get_snake_case(label)

    def __get_setter_name(self, key: str) -> str:
        """
        Returns the attributes setter name
        from its configuration key
        """

        setter_name = 'set_%s' % (self.__get_snake_case(key))
        return setter_name

    def initialize(self) -> None:
        """
        Performs inintialization operations
        on the configurable object
        """

        if not isinstance(self.configuration, dict):
            _config_notify.warning('Failed to initialize %s. Configuration is not a valid type (%s) expected dict' % (
                self.__class__.__name__, self.configuration.__class__.__name__))
            
            return

        for key, value in list(self.configuration.items()):
            setter_name = self.__get_setter_name(key)
            setter = getattr(self, setter_name, None)

            if setter:
                self.notify.debug('(Setter) %s: %s' % (setter_name, str(value)))
                if isinstance(value, tuple):
                    setter(*value)
                else:
                    setter(value)
            elif self.auto_configure:
                setattr(self, key, value)
            else:
                _config_notify.warning('%s does not implement setter: %s' % (
                    self.__class__.__name__, setter_name))

        self.configuration = {}
        self.__config = None
        self._included_config = None

    def load_data(self, section: str, data: object) -> None:
        """
        Processes the incoming data from a config section. 
        Intended to be overridden by child objects
        """

        if self.__want_config_warnings:
            _config_notify.warning('%s does not implement general section loader "load_data"' % (
                self.__class__.__name__))

    def pop(self, attr: str, default: object = None) -> object:
        """
        Pops the requested attribute from the 
        configuration dictionary
        """

        return self.configuration.pop(attr, default)

    def pop_as(self, attr: str, cls: object, default: object = None) -> object:
        """
        Pops the requested attribute from the 
        configuration dictionary and passes it as the first argument
        into the provided cls object
        """

        assert cls != None

        variable = self.pop(attr, default)
        assert variable != None

        return cls(variable)

    def pop_call(self, attr: str, func: object, default: object = None, **kwargs) -> object:
        """
        Pops the requested attribute from the configuration dictionary
        and passes it into the function callback returning the result
        """

        assert func != None
        assert callable(func)

        variable = self.pop(attr, default)
        assert variable != None

        return func(variable, **kwargs)

    def get(self, attr: str, default: object = None) -> object:
        """
        Retrieves the requested attribute from
        the configuration dictionary
        """

        return self.configuration.get(attr, default)

    def get_as(self, attr: str, cls: object, default: object = None) -> None:
        """
        Retrieves the requested attribute from the configuration
        dictionary and passes it as the first argument in the cls 
        object
        """

        assert cls != None

        variable = self.get(attr, default)
        assert variable != None

        return cls(variable)

    def get_call(self, attr: str, func: object, default: object = None, **kwargs) -> None:
        """
        Retrieves the requested attribute from the configuration
        dictionary and passes it as the first argument into the function 
        returning the results
        """

        assert func != None
        assert callable(func)

        variable = self.get(attr, default)
        assert variable != None

        return func(variable)        

    def reload(self) -> None:
        """
        Reloads the configuration object
        """

        self.load()
        self.initialize()



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys

from panda3d_gemstone.framework import exceptions

try:
    from PyQt5.QtWidgets import QApplication
except ImportError:
    raise exceptions.MissingThirdpartySupportError('PyQt5')

from panda3d_gemstone.application.application import GemstoneApplication

from panda3d_gemstone.editor.showbase import EditorShowbase
from panda3d_gemstone.editor.window import EditorWindow
from panda3d_gemstone.editor.styles import StyleManager
from panda3d_gemstone.editor.level import EditorWorldManager, EditorScene

from panda3d_gemstone.engine import runtime, prc
from panda3d_gemstone.framework.progress import RenderController

from panda3d.core import WindowProperties

class EditorQApplication(QApplication):
    """
    Custom QApplication instance for use inside the Gemstone Editor application
    """

    def __init__(self, application: object):
        self._application = application
        super().__init__(sys.argv)

    def setup(self) -> None:
        """
        Performs setup operations on the QApplication instnace
        """

        # Sync PyQT application information with Gemstone application
        # information.
        self.setApplicationDisplayName(self._application.get_application_name())
        self.setApplicationName(self._application.get_application_name())
        self.setApplicationVersion(self._application.get_application_version())

    def run(self) -> int:
        """
        Runs the QApplication instance
        """

        return self.exec_()

class EditorApplication(GemstoneApplication):
    """
    Base class for all Gemstone editor applications
    """

    SHOWBASE_CLS = EditorShowbase
    
    def __init__(self, showbase_cls: object = None):
        super().__init__(window_type='none', showbase_cls=showbase_cls)
        self.world_manager = EditorWorldManager.instantiate_singleton()
        self.editor_scene = EditorScene.instantiate_singleton()
        self.editor_controller = RenderController()
        self.q_application = EditorQApplication(self)
        self.editor_window = None

    def load_prc_files(self, files: list, optional: bool = False) -> None:
        """
        Attempts to load a list of prc files
        """

        if not len(files):
            return

        for prc_file in files:
            prc.load_prc_file(prc_file, optional)

    def setup(self) -> None:
        """
        Performs setup operations on the application instance
        """

        super().setup()

        # Setup PyQT components
        self.q_application.setup()
        self.styles = StyleManager(self)
        self.styles.set_dark_theme()

        # Create the editor window
        self.editor_window = EditorWindow(self)

    def run(self) -> int:
        """
        Runs the editor application
        """

        if not self.window:
            self.notify.error('Failed to start editor. Could not open window')
            return

        self.notify.info('Opening primary editor window.')
        self.editor_window.show()

        self.notify.info('Starting application loop.')
        return self.q_application.run()

    def get_application_name(self) -> str:
        """
        Returns the application's name. 
        """

        return '%s - Level Editor' % (
            prc.get_prc_string('app-name', 'Gemstone Application'))





"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging

from panda3d_gemstone.engine import runtime, prc

from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.runnable import Runnable
from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.service import Service, Command
from panda3d_gemstone.framework.controller import ControllerChainMixIn
from panda3d_gemstone.controllers.basic import FixedController

class Camera(Configurable, Runnable):
    """
    Represents the base class of all cameras in
    the Gemstone framework
    """

    def __init__(self, config_path: str):
        self.__name = ''
        self.__fov = prc.get_prc_double('default-fov', 30.0)
        self.__near = prc.get_prc_double('default-near', 1.0)
        self.__far = prc.get_prc_double('default-far', 100000)
        Configurable.__init__(self, config_path)
        Runnable.__init__(self, priority=10)
        CameraManager.get_singleton()._register(self)

    @property
    def camera_id(self) -> object:
        """
        Returns the camera's id as a property
        """

        return self.get_camera_id()

    @property
    def name(self) -> str:
        """
        Returns the camera's name as a property
        """

        return self.get_name()

    @name.setter
    def name(self, name: str) -> None:
        """
        Sets the camera's name as a setter
        """

        self.set_name(name)

    @property
    def camera(self) -> object:
        """
        Returns the showbase camera object
        """

        return runtime.camera

    @property
    def camera_lens(self) -> object:
        """
        Returns the showbase camera lens object
        """

        return runtime.base.camLens

    def get_id(self) -> object:
        """
        Returns this camera's id 
        """

        return id(self)

    def get_name(self) -> str:
        """
        Returns the cameras name
        """

        return self.__name
    
    def set_name(self, name: str) -> None:
        """
        Sets the camera's name
        """

        self.__name = name

    def destroy(self) -> None:
        """
        Called on the destruction of the camera
        """

        camera_mgr = CameraManager.get_singleton()
        if self in camera_mgr.get_cameras():
            cam_mgr._unregister(self)

    def activate(self) -> None:
        """
        Activates the camera object
        """

        if self.is_activated():
            return False

        activated = Runnable.activate(self)
        CameraManager.get_singleton()._activate_camera(self)

        self.set_fov(self.__fov)
        self.set_near(self.__near)
        self.set_far(self.__far)

        return activated

    def deactivate(self): 
        """
        Deactivates the camera object
        """

        if self.is_activated():
            CameraManager.get_singleton()._deactivate_camera(self)

        return Runnable.deactivate(self)

    def reload(self):
        """
        Reloads the camera object
        """

        self.load()
        self.initialize()

    def set_fov(self, fov: float) -> None:
        """
        """

        self.__fov = fov
        if self.camera_lens and self.is_activated():
            self.camera_lens.set_fov(fov)

    def get_fov(self) -> float:
        """
        """

        return self.__fov

    def set_far(self, far: float) -> None:
        """
        """

        self.__far = far
        if self.camera_lens and self.is_activated():
            self.camera_lens.set_far(far)

    def get_far(self) -> None:
        """
        """

        return self.__far

    def set_near(self, near: float) -> None:
        """
        """
    
        self.__near = near
        if self.camera_lens and self.is_activated():
            self.camera_lens.set_near(near)

    def get_near(self) -> None:
        """
        """

        return self.__near

    def __getattr__(self, name) -> object:
        """
        Custom attribute for briding the Gemstone camera object
        to the native Panda3D camera lens and camera object created
        by the application's showbase instance
        """

        base = runtime.base
        if hasattr(base.camLens, name):
            return getattr(base.camLens, name)
        elif hasattr(base.camera, name):
            return getattr(base.camera)
        else:
            raise AttributeError('%s does not have attribute: %s!' % (
                self.__class__.__name__, name))

    def __del__(self) -> None:
        """
        Calls destroy on object deletion
        """

        self.destroy()

class FixedCamera(Camera):
    """
    """

    def __init__(self, config_path: str, target: object):
        super().__init__(config_path)
        self.__controller = FixedController(config_path)
        self.__target = None
        self.set_target(target)
        self.initialize()

    @property
    def controller(self) -> object:
        """
        Returns this camera's controller
        """

        return self.__controller

    @property
    def target(self) -> object:
        """
        Returns this camera's target point as a
        property
        """

        return self.get_target()

    @target.setter
    def target(self, target: object) -> None:
        """
        Sets this camera's target point as a 
        setter
        """

        self.set_target(target)

    def destory(self) -> None:
        """
        Called on camera destruction
        """

        if self.__controller != None:
            self.__controller.destroy()
            self.__controller = None

        Camera.destroy(self)

    def set_target(self, target: object) -> None:
        """
        Set's this camera's target point
        """

        self.__target = target
        self.__controller.foreach_call('set_target', target)

    def get_target(self) -> object:
        """
        Returns this camera's target point
        """

        return self.__target

    async def tick(self, dt: float) -> None:
        """
        Performs camera tick operations once per frame
        """

        np = await self.__controller.update(camera, dt)
        self.camera.set_quat(np.get_quat())
        self.camera.set_fluid_pos(np.get_pos())

class FollowCamera(Camera, ControllerChainMixIn):
    """
    """

    def __init__(self, config_path: str, target: object):
        super().__init__(config_path)
        self.__target = None
        self.__target_pos = None
        self.__target_hpr = None
        self.__mouse_rotating_handling = True
        self.__update_target_look_at = False
        self.__controller = self.setup_controller_chain(self.pop('controller'), config_path)
        self.__mouse_controller = self.find_controller(self.controller, MouseRotationController)
        self.set_target(target)
        self.initialize()

    def set_mouse_rotation_handling(self, mouse_rotation_handling: bool) -> None:
        """
        """

        self.__mouse_rotating_handling = mouse_rotation_handling

    def get_mouse_rotation_handling(self) -> bool:
        """
        """

        return self.__mouse_rotating_handling

    def set_update_target_look_at(self, update_target_look_at: bool) -> None:
        """
        """

        self.__update_target_look_at = update_target_look_at

    def get_update_target_look_at(self) -> bool:
        """
        """

        return self.__update_target_look_at

    def set_target(self, target: object) -> None:
        """
        Sets the follow camera's target NodePath
        """

        self.__target = target
        self.__target_pos = target.get_pos()
        self.__target_hpr = target.get_hpr()
        self.call_controller_chain(self.__controller, 'set_target', target)

    def get_target(self) -> object:
        """
        Returns the follow camera's target NodePath
        """

        return self.__target

    def activate(self) -> bool:
        """
        Attempts to activate the follow camera
        """

        result = Camera.activate(self)
        if result:
            self.call_controller_chain(self.controller, 'activate')

        return result

    def deactivate(self) -> bool:
        """
        Attempts to adeactivate the follow camera
        """

        result = Camera.deactivate(self)
        if result:
            self.call_controller_chain(self.controller, 'aeactivate')

        return result

    async def tick(self, dt: float) -> None:
        """
        Called once per frame to perform tick operations on the FollowCamera
        object instance
        """

        # Verify we have a controller instance
        if self.__controller is None:
            return

        np = await self.__controller.update(self.__target, dt)
        if self.__mouse_controller and self.__mouse_rotating_handling:
            should_lerp = self.__target_pos.almost_equal(self.__target.get_pos()) or self.__target_hpr.almost_equal(self.__target.get_hpr())
            if should_lerp:
                self.__mouse_controller.set_mode(MouseRotationController.LERP_ROTATION)
            else:
                self.__mouse_controller.set_mode(MouseRotationController.SAVE_ROTATION)

        look_at = np.get_quat()
        position = np.get_pos()

        if camera:
            camera.set_quat(look_at)
            camera.set_fluid_pos(position)

        if self.__update_target_look_at:
            if self.__mouse_controller.dragging:
                if not self.__target.get_quat().almost_equal(look_at):
                    self.__mouse_controller._reset_position()
                    self.__target.set_quat(look_at)
            else:
                self.__mouse_controller._reset_angle()

        self.__target_pos = self.__target.get_pos()
        self.__target_hpr = self.__target.get_hpr()

class CameraManager(Singleton, Service):
    """
    """

    def __init__(self):
        Singleton.__init__(self)
        Service.__init__(self)
        self.__cameras = []
        self.__last_camera_index = 0
        self.__active_camera = None
        self.__last_active_camera = None
        self.add_command(Command('reload cameras', self.reload))
        self.add_command(Command('toggle active camera', self.toggle_active_camera))
        self.add_command(Command('toggle culling camera', self.toggle_culling_camera))
        self.add_command(Command('toggle oobe camera', self.toggle_oobe_camera))
        self.activate()

    @property
    def cameras(self) -> list:
        """
        Complete list of all cameras known by the CameraManager
        """

        return self.get_cameras()

    @property
    def active_camera(self) -> Camera:
        """
        Currently active camera instance
        """

        return self.get_active_camera()

    def has_camera(self, camera: Camera) -> bool:
        """
        Returns true if the camera is found in the camera manager instance
        """

        return camera in self.__cameras

    def _register(self, camera: Camera) -> None:
        """
        Registers a new camera object with the CameraManager singleton
        """

        # Verify the camera is not already registered
        if self.has_camera(camera):
            self.notify.warning('Tried to register already registered camera (%s)' % camera.get_name())
            return
        
        self.__cameras.append(camera)
        camera._Camera__id = self.__last_camera_index
        self.__last_camera_index += 1

        #self.add_command(Command('activate camera <%d>' % camera.get_id()))
        #self.add_command(Command('activate camera <%s>' % camera.get_name()))

    def _unregister(self, camera: Camera) -> None:
        """
        Unregistered a camera object from the CameraManager singleton
        """

        # Verify the camera is currently registered
        if not self.has_camera(camera):
            self.notify.warning('Tried to unregister an unknown camera (%s)' % camera.get_name())
            return

        self.remove_command('activate camera <%d>' % camera.get_id())
        self.remove_command(Command('activate camera <%s>' % camera.get_name()))

        if self.__active_camera == camera:
            camera.deactivate()

        self.__cameras.remove(camera)

    def _activate_camera(self, camera: Camera) -> bool:
        """
        Attempts to activate the requested camera and sets it as the current active camera.
        Returns true on success. False on no change or failure.
        """

        if not self.has_camera(camera):
            self.notify.warning('Tried to activate an unknown camera: %s' % camera.get_name())
            return False
        
        if self.__active_camera == camera:
            self.notify.warning('Tried to activate already active camera: %s' % camera.get_name())
            return False
        
        # Deactivate the existing active camera if any.
        if self.__active_camera:
            self.__active_camera.deactivate()

        runtime.base.disableMouse()
        self.__active_camera = camera
        self.__last_active_camera = None

        return True

    def _deactivate_camera(self, camera: Camera) -> bool:
        """
        Deactivates the requested camera. Removing it from the current active camera
        variable. Returns true on success. False on no change or failure.
        """

        if not self.has_camera(camera):
            self.notify.warning('Tried to deactivate an unknown camera: %s' % camera.get_name())
            return False
        
        if self.__active_camera != camera:
            self.notify.warning('Tried to deactivate a camera other then active: %s' % camera.get_name())
            return False

        self.__active_camera = None
        self.__last_active_camera = None
        runtime.base.enableMouse()

        return True

    def get_active_camera(self) -> object:
        """
        Returns the currently active camera
        """

        return self.__active_camera

    def get_cameras(self) -> list:
        """
        Returns all cameras owned by the camera manger singleton
        """

        return self.__cameras

    def reload(self) -> None:
        """
        Reloads the camera manger and its owning cameras
        """

        for camera in self:
            camera.reload()

    def toggle_active_camera(self) -> None:
        """
        Toggles the active camera
        """

        if self.last_active_camera:
            self.last_active_camera.activate()
        elif self.active_camera:
            last_active_camera = self.active_camera
            last_active_camera.deactivate()
            self.last_active_camera = last_active_camera

    def toggle_culling_camera(self) -> None:
        """
        Toggles the oobe culling camera
        """

        runtime.base.oobe_cull()

    def toggle_oobe_camera(self) -> None:
        """
        Toggles the oobe camera
        """

        runtime.base.oobe()

    def __iter__(self):
        """
        Customer iterator for iterating all cameras in the camera manager
        singleton
        """

        for camera in list(self.cameras):
            yield camera


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
import re

from panda3d_gemstone.logging.utilities import get_notify_category

from panda3d_gemstone.framework.internal_object import PandaBaseObject
from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.cast import cast
from panda3d_gemstone.framework.utilities import get_snake_case

from panda3d_gemstone.engine import runtime

__service_notify = get_notify_category('service')

class Command(object):
    """
    Represents a command that can be registered with a 
    Service object instance
    """

    def __init__(self, name: str, func: object, *args):
        self.name = name
        self.func = func
        self.args = list(args)

    def __call__(self, *args) -> None:
        """
        Custom call handler to execute our command
        handler function
        """

        self.func(*(list(self.args) + list(args)))

    def destroy(self) -> None:
        """
        Destroys the Command object instance
        freeing it from memory
        """

        self.name = None
        self.func = None
        self.args = []

class State(Command):
    """
    Represents a state based command for the Service
    objects to register
    """

    def __init__(self, name: str, *initial_value):
        super().__init__(self, name, self.set_state)
        self.args = []
        
        if len(initial_value) == 0:
            self.set_state(initial_value[0])
        else:
            self.set_state(initial_value)

    @property
    def value(self) -> object:
        """
        Returns this State objects current state
        """

        return self.get_state()

    def destroy(self) -> None:
        """
        Destroys the State object instance
        freeing it from memory
        """

        self.state = None
        Command.destroy(self)

    def set_state(self, state: object, *args) -> None:
        """
        Sets the state variable
        """

        self.state = state

    def get_state(self) -> object:
        """
        Returns the current state variable
        """

        return self.state

def register_command(*args):
    """
    Registers a service function as a command 
    """

    def decorate(f: object) -> None: 
        """
        Custom decorator for attaching the __events__ attribute
        to the class function for use by the Service's startup intializer
        """

        setattr(f, '__events__', args)

    return decorate

class Service(object):
    """
    Base class for all services. Allows the registering of commands inside a
    Gemstone application
    """

    def __init__(self, name: str = None, priority: int = 0, initialize_handle_attributes_as_commands: bool = True):
        self.__name = name if name else get_snake_case(self.__class__.__name__)
        self.__priority = priority
        self.__running = False
        self.__commands = {}

        if initialize_handle_attributes_as_commands:
            self.__initialize_commands()

        self.reset_commands_being_called()

    @property
    def service_name(self) -> str:
        """
        Returns the service's name as a property
        """

        return self.get_service_name()

    @service_name.setter
    def service_name(self, service_name: str) -> None:
        """
        Sets the service's name as a setter
        """

        self.set_service_name(service_name)

    @property
    def service_priority(self) -> int:
        """
        Returns the service's priority as a property
        """

        return self.get_service_priority()

    @service_priority.setter
    def service_priority(self, priority: int) -> None:
        """
        Sets the service's priority as a setter
        """

        self.set_service_priority(priority)

    def destroy(self) -> None:
        """
        Destroys the service object
        """

        self.deactivate()
        self.__deinitialize_commands()
    
        for command in list(self.__commands.keys()):
            self.remove_command(command)

        self.__commands = {}

    def get_commands_being_called(self) -> bool:
        """
        """

        return self.__commands_called

    def reset_commands_being_called(self) -> bool:
        """
        """

        self.__commands_called = False

    def set_commands_being_called(self) -> bool:
        """
        """

        self.__commands_called = True

    def __initialize_commands(self) -> None:
        """
        """

        service_manager = ServiceManager.get_singleton()
        command_methods = [(name, obj) for name, obj in inspect.getmembers(self) if inspect.ismethod(obj)]

        for name, method in command_methods:
            events = getattr(method, '__events__', None)
            if not events:
                continue

            self.add_command(Command(name, method))
            for event in events:
                if type(event) == type(()):
                    service_manager.register_event(event[0], self.get_service_name(), name, *event[1:])
                else:
                    service_manager.register_event(event, self.get_service_name(), name)

    def __deinitialize_commands(self) -> None:
        """
        """

        service_manager = ServiceManager.get_singleton()
        command_methods = [(name, obj) for name, obj in inspect.getmembers(self) if inspect.ismethod(obj)]

        for name, method in command_methods:
            events = getattr(method, '__events__', None)
            if not events:
                continue

            for event in events:
                if type(event) == type(()):
                    service_manager.unregister_event(event[0], name)
                else:
                    service_manager.unregister_event(event, name)        

    def get_service_name(self) -> str:
        """
        """

        return self.__name

    def set_service_name(self, name: str) -> None:
        """
        """

        self.__name = name

    get_name = get_service_name
    set_name = set_service_name

    def get_service_priority(self) -> int:
        """
        """

        return self.__priority

    def set_service_priority(self, priority: int) -> None:
        """
        """

        self.__priority = priority

    def add_command(self, command: Command, dev: bool = False) -> None:
        """
        Adds a new command handler to the service instance
        """

        # Check if we need to restrict this command based
        # on developer build status
        if dev and not runtime.is_developer_build():
            return

        if command.name in self.__commands:
            __service_notify.warning('command "%s" is already defined for "%s"!. Overriding handler' % (command.name, self.__class__.__name__))

        self.__commands[command.name] = command

    def remove_command(self, command_name: str) -> None:
        """
        Attempts to remove a command handler from the service
        instance
        """

        if command_name not in self.__commands:
            __service_notify.warning('Failed to remove command. Command "%s" is not defined' % (command_name))
            return

        self.__commands[command_name].destroy()
        del self.__commands[command_name]

    def get_command(self, command_name: str) -> Command:
        """
        """

        return self.__commands.get(command_name, None)

    def has_command(self, command_name: str) -> bool:
        """
        """

        return command_name in self.__commands

    def run_command(self, command_name: str, *args) -> bool:
        """
        Runs the requested command
        """

        if not self.__running:
            return False

        command = self.get_command(command_name)
        if not command:
            __service_notify.warning('run_command failed: %s is not a known command' % command_name)
            return False

        callable = not inspect.iscoroutinefunction(command)
        if not callable:
            __service_notify.warning('run_command failed: Command %s needs to be run async' % command_name)
            return False

        self.set_commands_being_called()
        command(*args)

    async def run_async_command(self, command_name: str, *args) -> bool:
        """
        Runs the requested command in async form
        """

        if not self.__running:
            return False

        command = self.get_command(command_name)
        if not command:
            __service_notify.warning('run_command failed: %s is not a known command' % command_name)
            return False

        awaitable = not inspect.iscoroutinefunction(command)
        self.set_commands_being_called()

        if awaitable:
            await command(*args)
        else:
            command(*args)

    def activate(self) -> bool:
        """
        Activates the service object
        """

        if self.__running:
            return False

        ServiceManager.get_singleton().activate(self)
        self.__running = True

        return True

    def deactivate(self) -> bool:
        """
        Deactivates the service object
        """

        if not self.__running:
            return False

        ServiceManager.get_singleton().deactivate(self)
        self.__running = False

        return True

    def is_activated(self) -> bool:
        """
        """

        return self.__running

    def get_state(self, command_name: str) -> object:
        """
        Retrieves the state from a command instance if present
        """

        if command_name not in self.__commands:
            __service_notify.warning('Failed to get state; Unknown state: %s' % command_name)
            return
        
        command = self.__commands[command_name]
        return command.get_state()

    def set_state(self, command_name: str, state: object) -> None:
        """
        """

        if command_name not in self.__commands:
            __service_notify.warning('Failed to set state; Unknown state: %s' % command_name)
            return

        command = self.__commands[command_name]
        command.set_state(state)        

    __getitem__ = get_state
    __setitem__ = set_state

    def __str__(self) -> str:
        """
        """

        mapping = ServiceManager.get_singleton().get_events(self)
        mapping.sort()

        s = ''
        for event, command_name in mapping:
            s += '%s: %s\n' % (command_name, event)

        return s

class ServiceManager(Singleton, PandaBaseObject):
    """
    Service manager for handling commands and events in the gemstone framework
    """

    ACTION_SERVICE_RE = '[^\t\n\r\x0c\x0b\\:]+'
    ACTION_NAME_RE = '[^\t\n\r\x0c\x0b\\(]+'
    ACTION_ARGS_RE = '[^\t\n\r\x0c\x0b\\)]+'
    ACTION_RE = re.compile('\\s*(' + ACTION_SERVICE_RE + ':)*(' + ACTION_NAME_RE + ')\\s*(?:\\((' + ACTION_ARGS_RE + ')\\))?\\s*:\\s*(\\S+)')

    def __init__(self):
        Singleton.__init__(self)
        PandaBaseObject.__init__(self)

        self.services = []
        self.events = {}
        self.__activated = True

    async def __select_highest_priority_services(self, command_name: str) -> list:
        """
        """

        services = {}
        for service in self.services:
            if service.has_command(command_name):
                priority = service.get_service_priority()
                if priority in services:
                    services[priority].append(service)
                else:
                    services[priority] = [service]

        if not services:
            self.notify.warning('Failed to select highest priority services. No services found')
            return []
        
        priorities = list(services.keys())
        priorities.sort(reverse=True)
        return services[priorities[0]]

    async def __dispatch(self, event: object, *event_args) -> None:
        """
        Receives an incoming event from the Panda3D messenger and redirects
        it into a Gemstone service command to dispatch
        """

        if not self.__activated:
            self.notify.warning('Attempting to dispatch event while not active (%s)' % event)
            return

        commands = self.events[event]
        for service_name, command_name, args in commands:
            await self.dispatch(command_name, service_name, *(event_args + args))

    async def dispatch(self, command_name: str, service_name: str = None, *args) -> None:
        """
        Dispatches a command to the requested service if available. Otherwise 
        dispatches to all services in the application if they consume the fired
        command
        """

        services = await self.__select_highest_priority_services(command_name)
        if not len(services):
            self.notify.warning('Failed to dispatch command "%s". No services found' % (
                command_name))
            
            return

        for service in services:

            if service_name and service_name != service.get_service_name():
                continue

            command = service.get_command(command_name)
            if command:
                service.set_commands_being_called()

                awaitable = inspect.iscoroutinefunction(command)
                if awaitable:
                    await command(*args)
                else:
                    command(*args)

    def register_event(self, event: str, service_name: str, command_name: str, *args) -> None:
        """
        Registers an event with the service manager
        """

        if event not in self.events:
            self.events[event] = []
            self.notify.debug('Registering event "%s" to command "%s" from service "%s"' % (
                event, command_name, service_name))
            self.accept(event, self.__dispatch, [event])

        self.events[event].append((service_name, command_name, args))

    def unregister_event(self, event: str, service_name: str, command_name: str) -> None:
        """
        Unregisters an event from the service manager
        """

        if command_name is None:
            del self.events[event]
        else:
            rem = []

            for i, info in enumerate(self.events[event]):
                if info[1] == command_name:
                    rem.append(i)

            for i in rem.__reversed__():
                del self.events[event][i]

    def get_events(self, service: Service = None) -> list:
        """
        Returns all registered events within the service manager
        """

        if service:
            return [(event, command_name) for event, data in list(self.events.items()) for service_name, command_name, args in data if service_name == service.get_service_name()]
        else:
            return list(self.events.keys())

    def activate(self, service: Service = None) -> None:
        """
        Activates a service with the ServiceManager singleton
        """

        if not service:
            self.__activated = True
            return

        if service not in self.services:
            self.services.append(service)

    def deactivate(self, service: Service = None) -> None:
        """
        Deactivates a service from the ServiceManager singleton
        """

        if not service:
            self.__activated = False
            return

        if service not in self.services:
            self.services.remove(service)

    def parse(self, lst: object, path: str = None) -> None:
        """
        """

        def syntax_error(line: str, line_nr: int) -> None:
            """
            Logs a syntax error
            """

            if path != None:
                self.notify.warning('Syntax error "%s" (#%d) in file: %s' % (line, line_nr, path))
            else:
                self.notify.warning('Syntax error "%s" (#%d)' % (line, line_nr))

        line_nr = 0
        for line in lst:
            line_nr += 1
            stripped_line = line.strip()
            if stripped_line and stripped_line[0] != '#':
                m = ServiceManager.ACTION_RE.match(line)
                if not m:
                    syntax_error(line, line_nr)
                    continue
            
                args = []
                service_name, command_name, command_args, event = m.groups()
                if service_name:
                    service_name = service_name[:-1]
                
                if command_args:
                    args = list(map(cast, command_args.split(',')))
                
                self.register_event(event, service_name, command_name, *args)

    def load(self, path: str) -> None:
        """
        Opens an event map ini config file from the application's VFS and loads it
        into the service manager instance
        """

        self.parse(open(path, 'r'), path=path)

    @classmethod
    def analyze_services(cls) -> None:
        """
        Analyzes the application services. Printing debug
        information to the console
        """

        singleton = cls.get_singleton()

        print('--------------------------------------------------------')
        print(str(singleton))
        print('--------------------------------------------------------')

    def __str__(self) -> str:
        """
        Custom string handler for returning a debug string about the ServerManager singleton
        for debugging purposes
        """

        s = ''
        s += 'Active Services\n'
        s += 'Class Instance\n--------------------------------------------------------\n'
        for service in self.services:
            cls_name = service.__class__.__name__
            name = ''
            if hasattr(service, 'get_service_name'):
                name = service.get_service_name()
            s += '%s%s%s\n' % (service.__class__.__name__, ' ' * (30 - len(cls_name)), name)

        s += '\nActive Event Mapping\n'
        s += 'Event Service Command -> Call\n--------------------------------------------------------\n'
        events = list(self.events.keys())
        events.sort()

        for event in events:
            s += '%s' % event
            first_line = True
            for service_name, command_name, args in self.events[event]:
                for service in self.services:
                    if service_name and service_name != service.get_service_name():
                        continue

                    command = service.get_command(command_name)
                    if not command:
                        continue
                    
                    service_name = service.__class__.__name__
                    if hasattr(service, 'get_service_name') and service.get_service_name():
                        service_name = service.get_service_name()

                    if first_line:
                        first_line = False
                        s += ' ' * (30 - len(event))
                    else:
                        s += ' ' * 30

                    func = command.func
                    if hasattr(func, 'im_class'):
                        cls = command.func.__self__.__class__.__name__ + '.'
                    else:
                        cls = ''
                    
                    s += '%s %s -> %s%s%s\n' % (
                        service_name, command.name, cls, command.func.__name__, str(tuple(command.args) + tuple(args)))

                if first_line:
                    s += ' (unassigned)%s%s%s\n' % (' ' * (30 - len(event) - len(' (unassigned)')), command_name, str(tuple(args)))

        return s


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import logging
import datetime
from math import tan, radians

from panda3d.core import Texture, Vec2, Vec3, Thread, Filename
from panda3d.core import InternalName, PNMImage, BoundingBox, Point3
from panda3d.core import TextureAttrib

from panda3d_gemstone.io.file_system import fix_path
from panda3d_gemstone.logging import utilities as logging

__utility_notify = logging.get_notify_category('model-utilities')

def pnms_equal(lhs, rhs) -> bool:
    """
    """

def get_all_geom_nodes(model: object):
    """
    """

    for geom_node_path in model.find_all_matches('**/+GeomNode'):
        yield geom_node_path.node()

def get_all_geoms(model: object, mutable: bool = False):
    """
    """

    for geom_node in get_all_geom_nodes(model):
        num_geoms = geom_node.get_num_geoms()
        for i in range(num_geoms):
            if mutable:
                yield geom_node.modify_geom(index)
            else:
                yield geom_node.get_geom(index)

def get_all_geom_states(model: object):
    """
    """

    for geom in get_all_geom_nodes(model):
        for i in range(geom.get_num_geoms()):
            yield geom.get_geom_state(i)

def has_texture_alpha(texture: object) -> object:
    """
    """

    texture_format = texture.get_format()
    return textureFormat in [
        Texture.FAlpha,
        Texture.FRgba,
        Texture.FRgba4,
        Texture.FRgba5,
        Texture.FRgba8,
        Texture.FRgba12,
        Texture.FRgba16,
        Texture.FRgba32,
        Texture.FLuminanceAlpha,
        Texture.FLuminanceAlphamask
    ]

def is_model_transparent(model: object) -> bool:
    """
    """

def get_model_texture_objects(model: object) -> list:
    """
    """

    assert model != None

    texture = []
    textures_rs = model.find_all_textures()
    for i in range(textures_rs.get_num_textures()):
        textures.append(textures_rs.get_texture(i))

    overrides = [child.get_texture() for child in model.find_all_matches('**') if child.has_texture()]
    textures.extend(overrides)

    return textures


def get_model_textures(model: object) -> list:
    """
    """

    assert model != None
    return [fix_path(texture.get_filename().to_os_specific()) for texture in get_model_texture_objects(model)]

def save_model_as_maya_file(model: object, path: str, version: str = '2019', cleanup: bool = True) -> None:
    """
    Saves the model NodePath as a Maya file for editing and animation
    """

    assert model != None
    model.write_to_bam_file(path + '.bam')

    # Convert bam to egg
    __utility_notify.info('Converting %s.bam to %s.egg...' % (path, path))
    os.system('bam2egg %s.bam -o %.egg' % (path, path))

    # Convert egg to maya
    maya_executable = 'egg2maya%s' % version
    __utility_notify.info('Converting %s.egg to %s.mb using Maya %s...' % (path, path, version))
    os.system('%s %s.egg -o %s.mb' % (maya_executable, path, path))

    # Perform cleanup
    if cleanup:
        os.remove('%s.bam' % path)
        os.remove('%s.egg' % path)
        
def save_character(character: object, directory: str= 'charactershots', format: str = 'png') -> None:
    """
    """

def bake_anim_frame(actor: object, anim: object = None, frame: object = None) -> None:
    """
    """

def focus_model(model: object, camera: object = None, axis: Vec3 = Vec3(1, 1, 1), fix_planes: bool = False, zoom: float = 1.5) -> None:
    """
    """




"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import copy
import traceback

import glob

from panda3d.core import NodePath, Point3

from panda3d_gemstone.io.file_system import get_file_size, get_matching_files, path_exists
from panda3d_gemstone.framework.cast import cast
from panda3d_gemstone.framework.progress import ProgressCounter
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.utilities import disallow_production

from panda3d_gemstone.logging.utilities import get_notify_category

from panda3d_gemstone.engine.messenger import MessageListener, event_handler
from panda3d_gemstone.engine import runtime, prc
from panda3d_gemstone.world.entity import Entity, EntityProxy


__world_notify = get_notify_category('world')

class Layer(InternalObject):
    """
    Represents a layer for containing and managing entity objects in a Gemstone world
    """

    def __init__(self, id, world):
        super().__init__()
        self._id = id
        self._world = world
        self._entites = []
        self._hidden = False

    @property
    def id(self) -> str:
        """
        Returns this layer's id as a property
        """

        return self.get_id()

    @property
    def world(self) -> object:
        """
        Returns this layer's world as a property
        """
        
        return self.get_world()

    @property
    def hidden(self) -> bool:
        """
        Returns this layer's hidden state as a property
        """

        return self.is_hidden()

    def has_entity(self, entity: Entity) -> bool:
        """
        Returns true if the layer contains the requested
        entity instance
        """

        return entity in self._entites

    def add_entity(self, entity: Entity) -> None:
        """
        Adds a new entity to the layer
        """

        # Verify the layer does not already contain 
        # the requested entity
        if self.has_entity(entity):
            return

        self._entites.append(entity)

        # Hide the entity if we are currently hidden
        if self._hidden:
            entity.hide()

    def remove_entity(self, entity: Entity) -> None:
        """
        Removes the requested entity from the layer
        """
        
        # Verify the layer contains the requested entity
        if self.has_entity(entity):
            self.notify.warning('Attempted to remove nonexistant entity from layer %d' % (id(self)))
            return

        entity.remove_node()

    def _remove_entity(self, entity: Entity) -> None:
        """
        Removes the requested entity from the entity
        list
        """

        self._entites.remove(entity)

    def clear(self) -> None:
        """
        Clears the entity layer of all data
        """

        while self._entites:
            self._entites[-1].destroy()

        self._world = None
        self._entites = []

    def get_id(self) -> str:
        """
        Returns this layer's id
        """

        return self._id

    def get_world(self) -> object:
        """
        Returns this layer's world instance
        """

        return self._world

    def get_entity(self, name: str) -> object:
        """
        Attempts to retrieve entities from the layer
        that match the requested name
        """

        entities = []
        for entity in self._entites:
            if entity.get_name() == name:
                entities.append(entity)

        if len(entities) == 1:
            return entities[0]
        return entities

    def get_entities(self, filter: object = None) -> list:
        """
        Returns the layers entity list with an optional filter operation
        to perform on the retrieval
        """

        if filter is None:
            return copy.copy(self._entites)
        else:
            return [entity for entity in self._entites if list(filter(entity))]

    def is_hidden(self) -> bool:
        """
        Returns this layer's hidden state
        """

        return self._hidden

    def show(self) -> None:
        """
        Changes the layers hidden state to show the layer
        """

        # Verify we need to do any operations
        if not self.is_hidden():
            return

        self._hidden = False
        self.get_entities(lambda entity: entity.show())

    def hide(self) -> None:
        """
        Changes the layers hidden state to hide the layer
        """

        # Verify we need to do any operations
        if self.is_hidden():
            return

        self._hidden = True
        self.get_entities(lambda entity: entity.hide())

    @disallow_production
    def reload_source(self) -> None:
        """
        Called when the source of the application is reloaded
        """

        self.notify.debug('Reloading world: %s' % self.get_id())
        for entity in self.get_entities():
            has_reload = hasattr(entity, 'attempt_reload')
            if has_reload:
                entity.attempt_reload()

    def __str__(self) -> str:
        """
        Custom string handler for the layer object
        """

        return '%s <%s> %s' % (self.__class__.__name__, self._id, str(self._hidden))

class World(NodePath, InternalObject):
    """
    Represents a game world in the gemstone framework. Containing all layers and entities
    used to perform visuals and actions within the level
    """

    def __init__(self):
        NodePath.__init__(self, '')
        InternalObject.__init__(self)
        self._layers = {}

    def get_entity(self, name: str, layer_id: str = None) -> object:
        """
        """

        if layer_id in self._layers:
            return self._layers[layer_id].get_entity(name)

        entities = []
        for layer in self.get_layers():
            entity = layer.get_entity(name)

            if type(entity) == type([]):
                entities += entity
            else:
                entities.append(entity)

        if len(entities) == 1:
            return entities[0]
        return entities

    def has_entity(self, entity: Entity) -> bool:
        """
        Returns true if the world contains the requested
        entity instance
        """

        for layer in self.get_layer():
            if layer.has_entity(entity):
                return True

        return False

    def get_entities(self, layer_id: str, filter: object = None) -> list:
        """
        Gets all entities in the requested layer using a filter lambda 
        operation
        """

        if layer_id in self._layers:
            return self._layers[layer_id].get_entities(filter)

        return []

    def get_all_entities(self, filter: object = None) -> list:
        """
        Retrieves all entities in the world using the requested filter
        lambda operation
        """

        return [ entity in layer in self.get_layers() for entity in layer.get_entities(filter)] 

    def get_all_layer_ids(self) -> list:
        """
        Returns all layer ids in the world
        """

        return list(self._layers.keys())

    def get_layers(self) -> list:
        """
        Returns all layers in the world
        """

        return list(self._layers.values())

    def has_layer_id(self, layer_id: str) -> bool:
        """
        Returns true if the layer id exists in the world
        """

        return layer_id in self._layers

    def get_layer(self, layer_id: str, create_if_needed: bool = False) -> Layer:
        """
        Attempts to retrieve a layer if it exists. Otherwise
        creates a new layer if create_if_needed is true and returns
        the new layer instance
        """

        if self.has_layer_id(layer_id):
            return self._layers[layer_id]
        else:
            if not create_if_needed:
                return None

            self._layers[layer_id] = Layer(layer_id, self)
            return self._layers[layer_id]

    def clear(self) -> None:
        """
        Clears the world of all data
        """

        for layer in self.get_layers():
            layer.clear()

        self._layers = {}

    def save(self, fh: object, progress_controller: object = ProgressCounter) -> None:
        """
        Saves the world to file using a file handle and progress controller
        instance
        """

        entities = self.get_all_entities()
        progress_controller.start(len(self._layers) + len(entities), 0)

        for layer_id, layer in list(self._layers.items()):
            if layer.get_entities():
                print(layer, file=fh)
            progress_controller.increase(1)
    
        for entity in entities:
            print(entity, file=fh)
            progress_controller.increase(1)

        progress_controller.finish()

    def _create_layer(self, layer_id: str, hidden: bool, line_nr: int) -> None:
        """
        """

        layer = self.get_layer(layer_id, create_if_needed=True)
        if layer:
            if hidden:
                layer.hide()
        else:
            logging.error('%s.load: Failed to create layer "%s" at line %d' % (
                self.__class__.__name__, layer_id, line_nr))

    def _create_entity(self, entity_proxy: EntityProxy, line_nr: int) -> Entity:
        """
        """

        try:
            return entity_proxy.instantiate(self, line_nr)
        except Exception as e:
            self.notify.warning('Failed to create entity on line %d (Error: %s)' % (line_nr, str(e)))
            if prc.get_prc_bool('gs-detailed-entity-errors', False):
                self.notify.warning(traceback.format_exc())
            
            return None

    def load(self, fh, progress_controller = ProgressCounter()) -> None:
        """
        """

        self.query(fh, self._create_layer, self._create_entity, progress_controller)

    def query(self, fh, for_each_layer = None, for_each_entity = None, progress_controller = ProgressCounter()) -> None:
        """
        """

        progress_controller.start(get_file_size(fh), 0)
        line_nr = 0

        try:
            for line in fh.readlines():
                line_nr += 1
                line = line[:-1]
                elements = line.split(' ')
                cls_name = elements[0]

                if cls_name == 'Layer':
                    layer_id = elements[1].strip(' <>')
                    hidden = elements[2]
                    if for_each_layer:
                        for_each_layer(layer_id, cast(hidden), line_nr)
                else:
                    entity_proxy = EntityProxy.from_string(line)
                    if entity_proxy:
                        if for_each_entity:
                            for_each_entity(entity_proxy, line_nr)
                    else:
                        self.notify.warning('Failed to parse entity at line #%d' % (self.__class__.__name__, line_nr))

                #progress_controller.set_current(fh.tell())
        except:
            self.notify.warning('Failed to parse world entities in file: %s' % fh.name)
            print(traceback.format_exc())

        progress_controller.finish()

    def is_hidden(self, layer_id: str) -> bool:
        """
        Returns true if the requested layer is hidden
        """

        layer = self.get_layer(layer_id)
        if layer:
            return layer.is_hidden()

        return False

    def show(self, layer_id: str = None) -> None:
        """
        Shows the specified layer id if provided. Otherwise
        shows all layers.
        """

        if layer_id is None:
            for layer in self.get_layers():
                layer.show()
        else:
            layer = self.get_layer(layer_id)
            if layer:
                layer.show()

    def hide(self, layer_id: str = None) -> None:
        """
        Hides the specified layer id if provided. Otherwise
        hides all layers
        """

        if layer_id is None:
            for layer in self.get_layers():
                layer.hide()
        else:
            layer = self.get_layer(layer_id)
            if layer:
                layer.hide()

    @disallow_production
    def reload_source(self) -> None:
        """
        Called on application source reload. Reloads owning layers
        """

        for layer in self.get_layers():
            layer.reload_source()

class WorldManager(InternalObject, MessageListener):
    """
    """

    def __init__(self):
        InternalObject.__init__(self)
        MessageListener.__init__(self)
        self.accept('gs-file-changed', self.__on_file_change)
        self._worlds = {}
        self._collision = None

    @property
    def worlds(self) -> dict:
        """
        Property for accessing the worlds dictionary
        """

        return self._worlds

    @property
    def collision(self) -> object:
        """
        Property for accessing the WorldManager collision node
        """

        return self._collision

    @staticmethod
    def encode_world_coords(pos: tuple) -> str:
        """
        Encodes a 3d coordinate for the world transform
        """

        try:
            max = pow(2, 15)
            x = int(pos[0])
            y = int(pos[1])

            if x > max - 1 or x < -max or y > max - 1 or y < -max:
                __world_notify.warning('WorldManager.encode_world_coords: World position (%s) is invalid for encoding!' % str(pos))
                return '-' * 8
            return '%04x%04x' % (x + max, y + max) 
        except:
            return pos

    @staticmethod
    def decode_world_coords(code: str)-> tuple:
        """
        Decodes a 3d world transform coordinate
        """

        try:
            max = pow(2, 15)
            x = '0x%s' % code[:4]
            y = '0x%s' % code[4:]

            return (int(x, 16) - max, int(y, 16) - max, 0)
        except:
            return code

    def create_world(self, idx: object) -> World:
        """
        Creates a new world instance
        """

        return World()

    def __get_index(self, pos: object):
        """
        """

        if isinstance(pos, str):
            return pos
        else:
            return (int(pos[0]), int(pos[1]), 0)

    def has_world(self, idx: object) -> bool:
        """
        Returns true if the world manager contains a world
        with the requested idx
        """

        return idx in self._worlds

    def get_world(self, pos) -> World:
        """
        """

        idx = self.__get_index(pos)
        if not self.has_world(idx):
            world = self.create_world(idx)

            # Reparent the world if the render has been defined
            if runtime.has_render():
                world.reparent_to(runtime.render)

            self._worlds[idx] = world

        return self._worlds[idx]

    def __getitem__(self, pos) -> World:
        """
        Custom get item handler for retrieving a world
        by its idx from the WorldManager
        """

        idx = self.__get_index(pos)
        if idx in self._worlds:
            return self._worlds[idx]

        return None

    def __len__(self) -> int:
        """
        Custom length handler for counting all the
        worlds within the WorldManager
        """

        all_empty = True

        for world in self._worlds:
            if len(world):
                all_empty = False
                break
        
        if all_empty:
            return 0
        else:
            return len(self._worlds)

    def __iter__(self) -> iter:
        """
        Custom iteration handler for iterating the worlds
        within the WorldManager
        """

        return iter(self._worlds)

    def keys(self) -> list:
        """
        Returns the world dictionary keys
        """

        return list(self._worlds.keys())

    def values(self) -> list:
        """
        Returns the world dictionary values
        """

        return list(self._worlds.values())

    def items(self) -> list:
        """
        Returns the world dictionary items
        """

        return list(self._worlds.items())

    @disallow_production
    def __on_file_change(self, file_path: str, ext: str) -> None:
        """
        Called on application file change. Informs all its owning worlds
        """
        
        for world in self.values():
            world.reload_source()

    def get_all_entities(self, filter: object = None) -> list:
        """
        Retrieves all entities in the world manager by 
        iterating through all the worlds
        """

        entities = []
        for world in self.values():
            enitities += world.get_all_entities(filter)

        return entities

    def has_entity(self, entity) -> bool:
        """
        Iterates through all the worlds in the WorldManager
        and returns true if the entity has been identified in a child
        world instance
        """

        for world in self.values():
            if world.has_entity(entity):
                return True

        return False

    def clear(self) -> None:
        """
        Clears the WorldManager instance of all worlds and data
        """

        for world in self.values():
            world.clear()
            world.remove_node()

        if self._collision:
            self._collision.remove_node()
            self._collision = None

        self._worlds = {}

    def stash(self) -> None:
        """
        Stashes all the worlds belonging to the WorldManager instance
        """

        for world in self.values():
            world.stash()

    def unstash(self) -> None:
        """
        Unstashes all the worlds belonging to the WorldManager instance
        """

        for world in self.values():
            world.unstash()

    def _iterate_worlds(self, path: str, progress_controller: object, process_func: object, func: object, *args) -> None:
        """
        Iterates over all level files in a directory and performs the process callback on each level
        file path
        """

        self.notify.debug('Iterating worlds at path: %s' % path)

        # Verify the process function is callable
        if not callable(process_func):
            raise Exception('Failed to iterate worlds at: %s. Process function is not callable' % path)

        # Verify the callback is callable
        if not callable(func):
            raise Exception('Failed to iterate worlds at: %s. Callback is not callable' % path)

        # Retrieve all worlds worlds and verify results were returned
        worlds = get_matching_files(path, '*.lvl')
        if not worlds:
            self.notify.warning('Failed to iterate worlds at: %s. No worlds found' % (path))
            return

        # Process all found worlds
        progress_controller.start(len(worlds), 0)
        for world_name in worlds:
            process_func(world_name, func, *args)
            progress_controller.increase(1)

        progress_controller.finish()

    def _process_world(self, world_name: str, func: object, *args) -> None:
        """
        """

        self.notify.debug('Processing world (%s) with function: %s' % (world_name, str(func)))

        # Verify our function is callable
        if not callable(func):
            self.notify.warning('Failed to process world %s. %s is not callable' % (
                world_name, str(func)))
            
            return

        world_coord = WorldManager.decode_world_coords(os.path.basename(world_name))[1:-4]
        world = self.get_world(world_coord)
        fh = open(world_name, 'r')
        func(world, fh, *args)
        fh.close()

    def load(self, path: str, process_world = World.load, progress_controller: object = ProgressCounter()) -> None:
        """
        Loads all levels at the requested path if any exist
        """

        self._iterate_worlds(path, progress_controller, self._process_world, process_world)

    def query(self, path: str, for_each_layer: object = None, for_each_entity: object = None, progress_counter: object = ProgressCounter()) -> None:
        """
        """

        self._iterate_worlds(path, progress_counter, self._process_world, World.query, for_each_layer, for_each_entity)

    def save(self, path: str, progress_counter: object = ProgressCounter()) -> None:
        """
        """

        self.notify.info('Saving WorldManager worlds to folder: %s' % path)
        progress_counter.start(len(self._worlds), 0)
        
        # Verify save directory exists
        if not os.path.exists(path):
            os.mkdir(path)

        for i, world_coord in enumerate(self._worlds):
            world = self.get_world(world_coord)
            
            if world.get_all_entities():
                output_path = os.path.join(path, 'b%s.lvl' % WorldManager.encode_world_coords(world_coord))
                fh = open(output_path, 'w')
                world.save(fh)
                fh.close()
            
            progress_counter.increase(1)
        
        progress_counter.finish()

    def clear_path(self, path: str) -> None:
        """
        """

        files = glob.glob(os.path.join(path, '*.lvl'))
        for file in files:
            os.unlink(file)

    def flatten(self, type: str = 'medium') -> None:
        """
        """

        for world in self.values():
            self.notify.debug('Flattening: %s' % world)

            if type == 'light':
                world.flatten_light()
            elif type == 'medium':
                world.flatten_medium()
            elif type == 'strong':
                world.flatten_strong()
            else:
                self.notify.warning('WorldManager.flatten Failed to flatten world. Unknown type: %s' % type)

    def load_collisions(self, collision_name) -> None:
        """
        """

        # Verify path eixsts
        if not path_exists(collision_name):
            return

        try:
            self._collision = runtime.loader.load_model(collision_name)
            if self._collision and runtime.has_render():
                self._collision.reparent_to(runtime.render)
        except:
            pass


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.controller import Controller, ControllerService
from panda3d_gemstone.framework.utilities import get_camel_case
from panda3d_gemstone.framework.exceptions import raise_not_implemented

class AnimationController(Controller):
    """
    Base class for all animation controllers inside Gemstone
    """

    def __init__(self, config_path: str, next_controller: object = None, section: str = 'Animation'):
        self.__game_object = None
        self.__current_animation = None
        self.__animation_states = {}
        self.__playing_emote = None
        self.__current_emote = None
        self.__next_animation = None
        self.__transition_duration = 0
        self.__transition_speed = 0.1
        super().__init__(config_path, next_controller, section)

    def destroy(self) -> None:
        """
        Destroys the animation controller instance
        """

        self.deactivate()
        self.__game_object = None
        super().destroy()

    def deactivate(self) -> None:
        """
        Deactivates the controller object
        """

        self.__current_animation = None
        self.__playing_emote = None

    def activate(self) -> None:
        """
        Activates the controller instance
        """

        # Play default idle animation if present
        if self.has_idle_animation():
            self.play_animation(self.get_idle_animation())

    def is_setup(self) -> bool:
        """
        Returns true if the controller has been setup
        """

        return self.__game_object != None

    def setup(self, game_object: object) -> None:
        """
        Performs setup operations for the controller instance. Takes
        an instnace of the game object to operate on
        """

        if self.is_setup():
            self.notify.warning('Attempting to setup already setup %s animation controller.' % (
                self.__class__.__name__))

            return

        self.__game_object = game_object

    def load_animation_states_data(self, data: dict) -> None:
        """
        Loads the animation states data into memory
        """

        self.__animation_states = data

    def has_animation_state(self, animation: str) -> bool:
        """
        Returns true if the controller has the requested animation state
        """

        return animation in self.__animation_states

    def get_animation_state(self, animation: str, default: str = None) -> str:
        """
        Retrieves the value of an animation state if available. Otherwise 
        returns default value
        """

        if not self.has_animation_state(animation):
            self.notify.warning('No animation state named: %s. Defaulting to "%s"' % (
                animation, default))

        return self.__animation_states.get(animation, default)

    def set_animation_state(self, animation: str, value: str) -> None:
        """
        Sets the animation state value
        """

        self.__animation_states[animation] = value

    def __set_animation_weight(self, animation: str, weight: int) -> None:
        """
        Sets an animations weight on the actor object
        """

        if weight > 1:
            weight = 0
        elif weight < 0:
            weight = 0

        self.notify.debug('(%s) %s = %s' % (self.__game_object.get_name(), animation, weight))
        self.__game_object.set_control_effect(animation, weight)

    def __getattr__(self, key: str) -> object:
        """
        Custom get attribute handler for allowing method access
        to the controller's animation states
        """

        result = None
        if len(key) > 4:
            type_name = key[:3]
            state_name = key[4:].replace('_animation', '')
            state_name = get_camel_case(state_name)

            if type_name == 'get':
                result = lambda x=None: self.get_animation_state(state_name, x)
            elif type_name == 'set':
                result = lambda x: self.set_animation_state(state_name, x)
            elif type_name == 'has':
                result = lambda: self.has_animation_state(state_name)
            elif len(key) > 5:
                type_name = key[:4]
                state_name = get_camel_case(key[5:])
                if type_name == 'play':
                    result = lambda x=False: self.__play_wrapper(state_name, x)

        if not result:
            raise AttributeError('%s does not have attribute %s' % (
                self.__class__.__name__, key))
        
        return result

    def __play_wrapper(self, state: str, *args, **kwargs) -> bool:
        """
        Wrapper to provide play_x() access to the play_animation function
        """

        state = getattr(self, 'get_%s_animation' % (state))()
        return self.play_animation(state, *args, **kwargs)

    def is_in_transition(self) -> bool:
        """
        Returns true if the controller is currently in an animation transition
        """

        return self.__next_animation != None

    def cancel_transition(self) -> None:
        """
        Cancels the current animation transition sequence
        """
        
        if not self.is_in_transition():
            return

        self.__game_object.stop(self.__next_animation)
        self.__set_animation_weight(self.__next_animation, 0)
        self.__next_animation = None

    def play_animation(self, animation: str, force: bool = False, instant: bool = False, transition_speed: int = 1) -> bool:
        """
        Plays the requested animation with optional force to play input, instant play, and transition speed parameters. 
        Returning true if the animation was played, otherwise false
        """

        if not self.is_setup():
            return False

        self.notify.debug('Playing animation %s on game object (%s)' % (
            animation, str(self.__game_object)))
        
        success = False
        if force or (self.__current_animation != animation and not self.is_in_transition()):
            self.__game_object.loop(animation)

            if instant or self.__current_animation == None:

                if self.__current_animation != None:
                    self.__game_object.stop(self.__current_animation)
                    self.__set_animation_weight(animation, 0)

                if self.__next_animation != None:
                    self.__game_object.stop(self.__next_animation)
                    self.__set_animation_weight(self.__next_animation, 0)
                    self.__next_animation = None
                
                self.__current_animation = animation
                self.__set_animation_weight(animation, 1)
            else:
                self.__set_animation_weight(animation, 0)
                self.__next_animation = animation
                self.__transition_duration = 1.0
                self.__transition_speed = transition_speed
        
            success = True

        return success

    def is_emoting(self) -> bool:
        """
        Returns true if the controller is currently emoting
        """

        return self.__playing_emote or self.__current_emote != None

    def can_play_emote(self, emote: str) -> bool:
        """
        Returns true if the controller is allowed to play the requested 
        emote. Intended to be overriden by children. 
        """

        return not self.is_emoting()

    def play_emote(self, emote: str) -> bool:
        """
        Attempts to play the requested emote. Returning true if the
        emote was played. Otherwise False
        """

        if not self.can_play_emote(emote):
            return False
        
        self.notify.debug('(%s) emote = %s' % (self.__game_object.get_name(), emote))
        self.__game_object.play(emote)
        self.__set_animation_weight(emote, 1)
        self.__playing_emote = self.__game_object.get_duration(emote)

        self.cancel_transition()
        self.__game_object.stop(self.__current_animation)
        self.__set_animation_weight(self.__current_animation, 0)

        return True

    async def update_node(self, nodepath: object, dt: float) -> object:
        """
        Performs animation update operations on the controller's
        node update loop.
        """

        await self.do_animation_logic(dt)
        await self.update_animation(dt)

        return nodepath

    async def do_animation_logic(self, dt: float) -> None:
        """
        Called once per frame to perform animation update logic. Intended
        to be overriden by children.
        """

    async def update_animation(self, dt: float) -> None:
        """
        Called once per frame to update the animation state. Intended
        to be overriden by children.
        """

        if self.__playing_emote != None:
            if self.__playing_emote - dt <= 0.0:
                self.play_idle(True)

                self.__game_object.stop(self.__current_emote)
                self.__set_animation_weight(self.__current_emote, 0)

                self.__current_emote = None
                self.__playing_emote = None
            else:
                self.__playing_emote -= dt
        elif self.__next_animation != None:
            remaining = self.__transition_duration
            self.__set_animation_weight(self.__current_animation, remaining)
            self.__set_animation_weight(self.__next_animation, 1.0 - remaining)
            self.__transition_duration -= self.__transition_speed * dt

            if self.__transition_duration <= 0:
                self.__set_animation_weight(self.__current_animation, 0)
                self.__set_animation_weight(self.__next_animation, 1)

                self.__game_object.stop(self.__current_animation)
                self.__current_animation = self.__next_animation
                self.__next_animation = None

class MovementController(ControllerService):
    """
    Base class for all character movement controllers inside Gemstone
    """

    def __init__(self, config_path: str, next_controller: object = None, section: str = None):
        self.__collision_engine = None
        self.__fall_callback = None
        self.__stand_callback = None
        super().__init__(config_path, next_controller, section)

    async def do_movement_logic(self, nodepath: object) -> None:
        """
        """

    async def update_move(self, nodepath: object, dt: float) -> None:
        """
        """

    async def update_node(self, nodepath: object, dt: float) -> object:
        """
        """

class AvatarController(MovementController, AnimationController):
    """
    Base class for all avatar controllers inside Gemstone
    """

    def __init__(self, config_path: str, next_controller: object = None, section: str = None):
        MovementController.__init__(self, config_path, next_controller, section)
        AnimationController.__init__(self, config_path, next_controller, section)

    def destroy(self) -> None:
        """
        Destroys the avatar controller instance
        """

        self.deactivate()

        MovementController.destroy(self)
        AnimationController.destroy(self)

    async def update_node(self, node_path: object, dt: float) -> object:
        """
        Called once per frame to update the avatar node instance
        """

        if self.get_commands_being_called():

            await self.do_movement_logic(node_path)
            await self.do_animation_logic()

            self.reset_commands_being_called()

        await self.update_move(node_path, dt)
        await self.update_animation(dt)

        return node_path

    def deactivate(self) -> None:
        """
        Deactivates the avatar controller
        """

        MovementController.deactivate(self)
        AnimationController.deactivate(self)

    def activate(self) -> None:
        """
        Activates the avatar controller
        """

        MovementController.activate(self)
        AnimationController.activate(self)



    



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.progress import ProgressController
from panda3d_gemstone.framework.service import Service
from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.singleton import Singleton

class LoadingScreenBase(Service, Configurable, Singleton, InternalObject):
    """
    Base class for all loading screens in the Gemstone framework
    """

    def __init__(self, config_path: str, service_name: str = 'LoadingScreen'):
        Service.__init__(self, service_name or self.__class__.__name__)
        Singleton.__init__(self)
        Configurable.__init__(self, config_path)
        InternalObject.__init__(self)

    def destroy(self) -> None:
        """
        Destroys the loading screen object instance
        """

        Service.destroy(self)
        InternalObject.destroy(self)

    def update_loading_progress(self, section: object, current: int, max: int = None) -> None:
        """
        Updates the loading screen's progress. Intended to be overridden by children
        """

        raise NotImplementedError('%s does not implement update_loading_progress!' % self.__class__.__name__)

class LoadingScreenProgressController(ProgressController):
    """
    Custom progress controller for progressing the Gemstone 
    framework loading screen objects
    """

    def __init__(self, section: object, logic_cls: object):
        assert logic_cls != None
        assert hasattr(logic_cls, 'get_singleton')

        self.__section = section
        self.__logic_cls = logic_cls

    @property
    def section(self) -> object:
        """
        Section of loading logic to update
        """

        return self.__section

    def update(self, current: int, max: int) -> None:
        """
        Called on progression to update the loading screen
        """

        loading_screen = self.__logic_cls.get_singleton()
        if not loading_screen:
            return
        
        loading_screen.update_loading_progress(self.__section, current, max)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import types
import os

from panda3d_gemstone.engine import runtime
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.registry import ClassRegistryMixIn
from panda3d_gemstone.framework.progress import ProgressController
from panda3d_gemstone.io.indented_file import parse

class Scene(InternalObject, ClassRegistryMixIn):
    """
    Represents a scene in the Gemstone framework
    """

    @staticmethod
    def insert_node(nodepath: object, parent: object) -> None:
        """
        Inserts a node into the parent if it meets requirements
        """

        if hasattr(nodepath, 'reparent_to'):
            nodepath.reparent_to(parent)

    @staticmethod
    def remove_node(nodepath: object) -> None:
        """
        Removes a node from the scene graph if it meets requirements
        """

        if hasattr(nodepath, 'deactivate'):
            nodepath.deactivate()
        if hasattr(nodepath, 'remove_node'):
            nodepath.remove_node()

    @staticmethod
    def show_node(nodepath: object) -> None:
        """
        Shows the node in the scene graph if it meets requirements
        """

        if hasattr(nodepath, 'show'):
            nodepath.show()

    @staticmethod
    def hide_node(nodepath: object) -> None:
        """
        Hides the node from the scene graph if it meets requirements
        """

        if hasattr(nodepath, 'hide'):
            nodepath.hide()

    @staticmethod
    def insert_empty(object: object, parent: object) -> None:
        """
        Empty insert function placeholder
        """

    def __init__(self, path: str = None, progress_controller: ProgressController = ProgressController()):
        InternalObject.__init__(self)
        self.__objects = []
        if path is not None:
            self.load(path, progress_controller)

    def destroy(self) -> None:
        """
        Destroys the scene freeing it from memory
        """

        InternalObject.destroy(self)
        self.clear()

    def load(self, path: str, progress_controller: ProgressController = ProgressController()) -> None:
        """
        Loads a scene file from the applications file system
        """

        assert path != None
        assert path != ''

        nodes = parse(path)
        progress_controller.set_max(len(nodes))
        progress_controller.set_current(0)

        for node in nodes:
            progress_controller.tick()
            arg = node.args.strip()
            cls = self._get_class(node.type)
            if not cls:
                self.notify.warning('Failed to load object. Unkonwn class %s' % node.type)
                continue

            insert = self._get_class_meta(node.type, 'Scene.insert')
            if not insert:
                insert = Scene.insert_node
            
            if node.free:
                insert = None

            if arg:
                inst = cls(arg)
            else:
                inst = cls()

            node.instance = inst
            if node.name:
                self.__dict__[node.name] = inst
                if hasattr(inst, 'set_name'):
                    inst.set_name(node.name)
                elif hasattr(inst, 'setName'):
                    inst.setName(node.name)

            self.__objects.append(inst)
            if insert:
                if node.parent:
                    insert(inst, node.parent.instance)
                else:
                    insert(inst, runtime.render)

        progress_controller.finish()

    def get_object(self, name: str) -> None:
        """
        Attempts to retrieve an object by its name
        otherwise returns NoneType if not found
        """

        return getattr(self, name, None)

    def has_object(self, name: str) -> bool:
        """
        Returns true if the scene contains the object 
        name
        """

        return self.get_object(name) != None

    def get_objects(self, cls: object = None) -> list:
        """
        Retrieves all objects by their class if provided. Otherwise
        returns all objects
        """

        if not cls:
            return self.__objects

        objects = []
        if isintance(cls, bytes):
            for obj in self.__objects:
                if obj.__class__.__name__ == cls:
                    objects.append(obj)

        elif isinstance(cls, type):
            for obj in self.__objects:
                if obj.__class__ == cls:
                    objects.append(obj)

        return objects

    def __get_object_meta(self, obj: object, key: str, default: object = None) -> object:
        """
        Retrieves the object's class metadata information
        """

        meta = self._get_class_meta(obj.__class__.__name__, key)
        if not meta:
            meta = default

        return meta

    def __call_object_meta_func(self, obj: object, key: str, default: object = None) -> None:
        """
        Attempts to call an object's meta class function if it exists. Otherwise default
        """

        func = self.__get_object_meta(obj, key, default=default)
        if func:
            func(obj)

    def __call_object_meta_func_all(self, key: str, default: object = None) -> None:
        """
        Attempts to call the meta class function if it exists on all objects in the scene
        """

        for obj in self.__objects:
            self.__call_object_meta_func(obj, key, default)

    def clear(self) -> None:
        """
        Clears all objects from the scene
        """
        
        self.__call_object_meta_func_all('Scene.remove', default=Scene.remove_node)
        self.__objects = []

    def show(self) -> None:
        """
        Shows the scenes objects
        """

        self.__call_object_meta_func_all('Scene.show', default=Scene.show_node)

    def hide(self) -> None:
        """
        Hides the scenes objects
        """

        self.__call_object_meta_func_all('Scene.hide', default=Scene.hide_node)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import weakref
import traceback

from panda3d.core import PNMImage, Filename, TextureAttrib
from panda3d.core import Texture

from panda3d_gemstone.logging.utilities import get_notify_category
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.utilities import get_camel_case
from panda3d_gemstone.engine import runtime

__texture_notify = get_notify_category('texture')

def load_texture(*args, **kwargs) -> object:
    """
    Loads a texture using the standard Panda3D Loader
    method
    """

    return runtime.loader.load_texture(*args, **kwargs)

async def load_texture_async(*args, **kwargs) -> object:
    """
    Loads a texture using the standard Panda3D Loader
    method in an async way
    """

    return await runtime.loader.load_texture(*args, **kwargs)

def load_texture_as_buffer(self, file_path: str) -> object:
    """
    Loads a texture into a Gemstone TextureBuffer object
    """

    return TextureBuffer.load_file(file_path)

def load_heightfield(file_path: str) -> object:
    """
    Loads a heightfield from the application's VFS
    """

    try:
        return HeightfieldBuffer.load_file(file_path)
    except Exception as e:
        __texture_notify.warning('Failed to load heightfield (%s) as a buffer. Loading default texture' % (
            file_path))
        __terrain_notify.warning(traceback.format_exc())
        
        return load_texture(file_path)

def assign_texture_to_geom_node(geom_node: object, texture: object) -> None:
    """
    Assigns a texture to a GeomNode instance
    """

    if not texture:
        __texture_notify.warning('Failed to assign texture. Texture is NoneType')
        return

    assert geom_node != None
    assert hasattr(geom_node, 'get_num_geoms')

    num_geoms = geom_node.get_num_geoms()
    for index in range(num_geoms):
        geom = geom_node.get_geom(index)
        geom_state = geom_node.get_geom_state(index)
        
        if geom_state.get_attrib(TextureAttrib.get_class_type()) != None:
            new_state = geom_state.remove_attrib(TextureAttrib.get_class_type())
        else:
            new_state = geom_state

        new_state = new_state.add_attrib(TextureAttrib.make(texture))
        geom_node.set_geom_state(index, new_state)

def get_texture_from_geom_node(geom_node: object) -> object:
    """
    Retrieves the texture from a GeomNode instance
    """

    assert geom_node != None
    assert hasattr(geom_node, 'get_num_geoms')

    if geom_node.get_num_geoms() == 0:
        return None

    geom_state = geom_node.get_geom_state(0)
    return geom_state.get_attrib(TextureAttrib.get_class_type()).get_texture()

class TextureCache(InternalObject):
    """
    """

    def __init__(self, load_async: bool = False):
        super().__init__()

        self.texture_dict = {}
        self.image_dict = {}
        self.weakrefs = set()
        self.requested_for_cpu = set()
        self.requested_for_gpu = set()
        self.requested_for_cpu_and_gpu = set()
        self.set_load_async(load_async)

    def set_load_async(self, load_async: bool) -> None:
        """
        """

        if load_async:
            self.__load_texture_for_cpu_func = self.__load_texture_for_cpu
            self.__load_texture_for_gpu_func = self.__load_texture_for_gpu
            self.__load_texture_for_cpu_and_gpu_func = self.__load_texture_for_cpu_and_gpu
        else:
            self.__load_texture_for_cpu_func = self.__load_sync_Texture_for_cpu
            self.__load_texture_for_gpu_func = self.__load_sync_texture_for_gpu
            self.__load_texture_for_cpu_and_gpu_func = self.__load_sync_texture_for_cpu_and_gpu

    def destroy(self) -> None:
        """
        Destroys the texture cache instance
        """

        InternalObject.destroy(self)

        self.texture_dict = {}
        self.image_dict = {}
        self.weakrefs = set()
        self.requested_for_cpu = set()
        self.requested_for_gpu = set()
        self.requested_for_cpu_and_gpu = set()

    def has_texture(self, texture_name: str) -> bool:
        """
        Returns true if the texture is in the cache
        """

        return texture_name in self.texture_dict

    def has_image(self, image_name: str) -> bool:
        """
        Returns true if the image is in the cache
        """

        return image_name in self.image_dict

    def __load_image(self, path: str, *args, **kwargs) -> object:
        """
        Loads an image using the Panda3D loader object
        and returns the result
        """

        return runtime.loader.load_texture(path, *args, **kwargs)

    def __async_load_image(self, path: str, *args, **kwargs) -> object:
        """
        Loads an image using the Panda3D loader object asynchronously
        """

        return runtime.loader.asyncLoadTexture(path, *args, **kwargs)

    def request_texture_cpu(self, texture_name: str, querying_obj: object) -> None:
        """
        """

        exists = self.has_image(image_name)
        self.__request_texture_generate(texture_name, querying_obj, exists, self.requested_for_cpu, self.__load_texture_for_cpu_func)

    def request_texture_gpu(self, texture_name: str, querying_obj: object) -> None:
        """
        """

        exists = self.has_image(image_name)
        self.__request_texture_generate(texture_name, querying_obj, exists, self.requested_for_gpu, self.__load_texture_for_gpu_func)

    def request_texture_cpu_and_gpu(self, texture_name: str, querying_obj: object) -> None:
        """
        """

        exists = self.has_image(image_name)
        self.__request_texture_generate(texture_name, querying_obj, exists, self.requested_for_cpu_and_gpu, self.__load_texture_for_cpu_and_gpu_func)

    def __request_texture_generic(self, texture_name: str, querying_obj: object, existing: bool, requested_set: object, load_function: object) -> None:
        """
        """

        if existing:
            querying_obj.notify_of_new_texture()
        else:
            self.__store_object(querying_obj)
            if texture_name not in requested_set:
                requested_set.update([texture_name])
                load_function(texture_name)

    def __load_texture_for_cpu(self, texture_name: str) -> None:
        """
        """

        self.notify.debug('Requesting CPU only texture: %s' % texture_name)
        image = PNMImage()
        runtime.loader.asyncPnmImageRead(image, Filename(texture_name), callback=self.__texture_loaded_for_cpu_only_async_callback, extraArgs=[texture_name, image])

    def __load_texture_for_gpu(self, texture_name: str) -> None:
        """
        """

        self.notify.debug('Requesting GPU only texture: %s' % texture_name)
        self.__async_load_image(texture_name, callback=self.__texture_loaded_for_gpu_only_async_callback, extraArgs=[texture_name])

    def __load_texture_for_cpu_and_gpu(self, texture_name: str) -> None:
        """
        """

        self.notify.debug('Requesting CPU and GPU texture: %s' % texture_name)
        self.__async_load_image(texture_name, callback=self.__texture_loaded_for_cpu_and_gpu_async_callback, extraArgs=[texture_name])

    def __load_sync_texture_for_cpu(self, texture_name: str) -> None:
        """
        """

        self.notify.debug('Requesting CPU only texture: %s' % texture_name)
        image = PNMImage()
        success = PNMImage.read(image, Filename(texture_name))
        self.__texture_loaded_for_cpu_only_async_callback(success, texture_name, image)

    def __load_sync_texture_for_gpu(self, texture_name: str) -> None:
        """
        """

        self.notify.debug('Requesting GPU only texture: %s' % texture_name)
        texture = self.__load_image(texture_name)
        self.__texture_loaded_for_gpu_only_async_callback(texture, texture_name)

    def __load_sync_texture_for_cpu_and_gpu(self, texture_name: str) -> None:
        """
        """

        self.notify.debug('Requesting CPU and GPU texture: %s' % texture_name)
        texture = self.__load_image(texture_name)
        self.__texture_loaded_for_cpu_and_gpu_async_callback(texture, texture_name)

    def __store_object(self, obj: object) -> None:
        """
        Stores the weakref reference to the object for later use
        """

        self.weakrefs.update([weakref.ref(obj)])

    def __notify_all_objects(self) -> None:
        """
        """

        list_of_refs_to_remove = []
        for obj in self.weakrefs:
            reference = obj()
            if reference is None:
                list_of_refs_to_remove.remove(obj)
            else:
                if reference.notify_of_new_texture():
                    list_of_refs_to_remove.append(obj)
        
        for obj in list_of_refs_to_remove:
            self.weakrefs.remove(obj)

    def __texture_loaded_for_cpu_only_async_callback(self, success: bool, texture_name: str, texture: object) -> None:
        """
        """

        self.notify.debug('Received CPU texture: %s' % texture_name)
        self.request_texture_cpu.remove(texture_name)
        self.image_dict[texture_name] = texture

        if not success:
            self.notify.warning('Async texture load for CPU failed (%s)' % texture_name)
        
        self.__notify_all_objects()

    def __texture_loaded_for_gpu_only_async_callback(self, texture: object, texture_name: object) -> None:
        """
        """

        self.notify.debug('Received GPU texture: %s' % texture_name)
        self.requested_for_gpu.remove(texture_name)
        self.texture_dict[texture_name] = texture

        if not texture:
            self.notify.warning('Async texture load for GPU failed (%s)' % texture_name)

        self.__notify_all_objects()

    def __texture_loaded_for_cpu_and_gpu_async_callback(self, texture: object, texture_name: object) -> None:
        """
        """

        self.notify.debug('Received CPU and GPU texture: %s' % texture_name)
        self.requested_for_cpu_and_gpu.remove(texture_name)
        self.texture_dict[texture_name] = texture_name

        if not texture:
            self.notify.warning('Async texture load for CPU and GPU failed (%s)' % texture_name)

        self.__notify_all_objects()

class TextureBuffer(Texture, InternalObject):
    """
    Represents an editable texture buffer inside Gemstone
    """

    def __init__(self, *args, **kwargs):
        Texture.__init__(self)
        InternalObject.__init__(self)
        self._pnm = PNMImage(*args, **kwargs)
        self.refresh()

    @classmethod
    def load_file(cls: object, file_path: str) -> object:
        """
        Loads an image file into the TextureBuffer instance
        """

        instance = cls(Filename(file_path))
        return instance

    def __wrapper(self, method, *args, **kwargs) -> object:
        """
        Wrapper for a PNMImage function to auto update our 
        texture instance
        """

        results = None
        try:
            result = method(*args, **kwargs)
        except Exception as e:
            raise e

        self.refresh()
        return results

    def refresh(self) -> None:
        """
        Refreshes the texture instance from the internal PNMImage
        object
        """

        self.notify.debug('Refreshing %s %d' % (
            self.__class__.__name__, id(self)))
        self.load(self._pnm)

    def __getattr__(self, key: str) -> object:
        """
        Custom attribute getter for wrapping the Panda3D PNMImage object
        """

        results = None
        is_getter = key.startswith('get')
        if hasattr(self._pnm, key):
            results = getattr(self._pnm, key)
            if is_getter:
                results = lambda *args, **kwargs: self.__wrapper(results, *args, **kwargs)
        elif hasattr(self._pnm, get_camel_case(key)):
            results = getattr(self._pnm, get_camel_case(key))
            if is_getter:
                results = lambda *args, **kwargs: self.__wrapper(results, *args, **kwargs)

        if not results:
            raise AttributeError('%s does not have attribute %s' % (
                self.__class__.__name__, key))

        return results

class HeightfieldBuffer(TextureBuffer):
    """
    A Heightfield specific variant of the TextureBuffer. Automatically converts
    the incoming texture into a valid heightfield for use with GPUTerrain objects
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__set_max_value()

    def __set_max_value(self) -> None:
        """
        Sets the required max value of a gpu heightfield
        """
        
        self.set_maxval(65535)
        


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.engine import runtime

class ProgressCounter(InternalObject):
    """
    """

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.start(*args, **kwargs)

    @property
    def max(self) -> float:
        """
        Returns this counters max as a property
        """

        return self.get_max()

    @max.setter
    def max(self, max: float) -> None:
        """
        Sets this counters max a setter
        """

        self.set_max(max)

    @property
    def current(self) -> float:
        """
        Returns this counters current as a property
        """

        return self.get_current()

    @current.setter
    def current(self, current: float) -> None:
        """
        Sets this counters current as a setter
        """

        self.set_current(current)

    def start(self, max: float = 0, current: float = 0) -> None:
        """
        """

        self._max = max
        self._current = current

    def get_max(self) -> float:
        """
        """

        return self._max

    def set_max(self, max: float) -> None:
        """
        """

        self._max = max

    def get_current(self) -> float:
        """
        """

        return self._current

    def set_current(self, current: float) -> None:
        """
        """

        self._current = current

    def reset_current(self) -> None:
        """
        Reset's this counters current value to zero
        """

        self.set_current(0)

    def finish(self):
        """
        Completes the counter by setting its current to the
        configured max value
        """

        self.set_current(self._max)

    def increase(self, value: float) -> None:
        """
        """

        current = self._current + value
        self.set_current(min(current, self._max))

    def decrease(self, value: float) -> None:
        """
        """

        current = self._current + value
        self.set_current(max(current, 0))

    def tick(self) -> None:
        """
        Increaes the counter by one
        """

        self.increase(1)

class ProgressController(ProgressCounter):
    """
    """

    @property
    def update_frequency(self) -> float:
        """
        Returns this controllers update frequency as a property
        """

        return self.get_update_frequency()

    @update_frequency.setter
    def update_frequency(self, frequency: float) -> None:
        """
        Sets this controllers update frequency as a setter
        """

        self.set_update_frequency(frequency)

    def start(self, max: float = 0, current: float = 0, update_frequency: float = 0.0) -> None:
        """
        """

        ProgressCounter.start(self, max, current)
        self._update_frequency = update_frequency
        self._last_update = 0.0

    def get_update_frequency(self) -> float:
        """
        Returns this controller's update frequency
        """

        return self._update_frequency

    def set_update_frequency(self, update_frequency: float) -> None:
        """
        Sets this controller's update frequency
        """

        self._update_frequency = update_frequency
        self._last_update = 0.0

    def set_current(self, current: float) -> None:
        """
        """

        ProgressCounter.set_current(self, current)
        dt = globalClock.get_dt()

        self._last_update += dt
        if self._last_update > self._update_frequency:
            self._last_update = 0
            
            sig = inspect.getfullargspec(self.update)
            args = len(sig[0])
            if args == 1:
                self.update()
            elif args == 2:
                self.update(current)
            elif args > 2:
                self.update(current. self.get_max())

    def update(self) -> None:
        """
        Called on progress counter change
        """

        raise NotImplementedError('%s does not implement update!' % self.__class__.__name__)

class RenderController(ProgressController):
    """
    Custom Progress Controller instance that forces
    a frame render each progress update
    """

    def update(self, current: object) -> None:
        """
        Called on progress counter change. Instructs
        the graphics engine to render a frame
        """

        if not runtime.has_base():
            self.notify.warning('Attempted to render frame via %s without a ShowBase!' % self.__class__.__name__)
            return

        try:
            runtime.base.graphicsEngine.render_frame()
        except:
            self.notify.warning('Failed to render frame for %s update: %s' % (self.__class__.__name__, current))


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import os

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.exceptions import raise_not_implemented

from panda3d_gemstone.logging.utilities import get_log_directory, get_log_filename, get_notify_category

class EventHandler(InternalObject):
    """
    Base class for all Gemstone event handlers
    """

    def __init__(self):
        super().__init__()
        self.setup()

    def __del__(self, *args, **kwargs) -> None:
        """
        Called on object deletion
        """

        self.destroy()

    def setup(self) -> None:
        """
        Called on event handler setup
        """

    def destroy(self) -> None:
        """
        Called on event handler destruction
        """

        super().destroy()

    def consume_event(self, type=None, **data) -> None:
        """
        Consumes the incoming event from the EventsManager
        """

        raise_not_implemented(self, 'consume_event')

class ConsoleEventHandler(EventHandler):
    """
    Console printout event handler for Gemstone
    """

    def consume_event(self, type=None, **data) -> None:
        """
        Consumes the incoming event from the EventsManager
        """

        if not hasattr(self, '_notify'):
            self._notify = get_notify_category('events')
            self._notify.setInfo(True)

        event_message = '%s: %s' % (type, json.dumps(data))
        self._notify.info(event_message)

class FileEventHandler(EventHandler):
    """
    File base file handler for Gemstone
    """

    def setup(self) -> None:
        """
        Performs setup operations on the FileEventHandler
        """

        log_directory = get_log_directory()
        log_filename = 'events_' + get_log_filename()
        if not os.path.exists(log_directory):
            os.mkdir(log_directory)

        log_path = os.path.join(log_directory, log_filename)
        self.notify.debug('Opening event log: %s' % log_path)
        self._log_file = open(log_path, 'w')

    def destroy(self) -> None:
        """
        Destroys the FileEventHandler instance
        """

        if hasattr(self, '_log_file'):
            self._log_file.close()
            del self._log_file

    def consume_event(self, type=None, **data) -> None:
        """
        Consumes the incoming event from the EventsManager
        """

        assert hasattr(self, '_log_file')
        assert self._log_file != None

        message = '[%s]' % type
        parts = []
        for data_key, data_value in data.items():
            parts.append('%s: %s' % (data_key, data_value))

        message = '%s (%s)\n' % (message, ', '.join(parts))
        self.notify.debug('Logging Event: %s' % message)
        self._log_file.write(message)

class EventsManager(Singleton, InternalObject):
    """
    Handles all events in the Gemstone framework
    """

    def __init__(self):
        Singleton.__init__(self)
        InternalObject.__init__(self)
        self.notify.setInfo(True)
        self.__handlers = []
        self.add_handler(ConsoleEventHandler())

    def destroy(self) -> None:
        """
        Called on the destruction of the events manager
        """

        Singleton.destroy(self)
        InternalObject.destroy(self)

    def add_handler(self, handler: EventHandler) -> bool:
        """
        Adds a new event handler to the EventsManager. Returning True on success
        """

        if handler in self.__handlers:
            self.notify.warning('Failed to add event handler %s. Handler already exists in EventsManager' % (
                handler.__class__.__name__))

            return False

        self.notify.info('Registering event handler: %s' % handler.__class__.__name__)
        self.__handlers.append(handler)

        return True

    def consume_event(self, type=None, **data) -> None:
        """
        Consumes an application event and routes it to the available handlers
        """

        # Verify we have handlers present
        if not len(self.__handlers):
            self.notify.warning('Failed to consume %s event. No handlers registered!' % type)
            return

        for handler in self.__handlers:
            handler.consume_event(type, **data)

def log_event(type='GenericEvent', **data) -> None:
    """
    Logs an application event with the EventsManager instance
    """

    em = EventsManager.get_singleton()
    if not em:
        return
    
    em.consume_event(type, **data)

def add_event_handler(*args, **kwargs) -> bool:
    """
    Adds an event handler to the EventsManager singleton. Returns
    true if the operation was a success
    """

    em = EventsManager.get_singleton()
    if not em:
        return False

    return em.add_handler(*args, **kwargs)

    
